"""
This module holds the Script class, which handles the generation and parameterization
of scripts.
"""

from __future__ import annotations

import copy
import logging
import os.path
import re
from os import PathLike
from typing import Union, Any, List

from remotemanager.connection.computers import Substitution
from remotemanager.connection.computers.dynamicvalue import (
    DynamicMixin,
    EMPTY_TREATMENT_STYLES,
    INVALID_EMPTY_TREATMENT,
)
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils.tokenizer import Tokenizer
from remotemanager.utils.uuid import generate_uuid

logger = logging.getLogger(__name__)

DELETION_FLAG_LINE = "~marked_for_line_deletion~"
DELETION_FLAG_LOCAL = "~marked_for_local_deletion~"
PLACEHOLDER_PATTERN = r"#(\w+)(?::([^#]+))?#"
REPLACEMENT_PATTERN = r"#{name}(?::[^#]+)?#"
ESCAPE_SEQ_PATTERN = r"(?<!!\\)\\(?!\\)"


class EscapeStub:
    """
    Stub class for avoiding regex's internal escape sequence handling

    If the `repl` argument of `re.sub` is a callable, escape sequences will not be
    processed, allowing us to handle them at a later stage.

    This is important for allowing templates to escape the `:` character
    """

    __slots__ = ["content"]

    def __init__(self, content: Any):
        self.content: str = str(content)

    def __call__(self, *args, **kwargs) -> str:
        return self.content

    def __str__(self) -> str:
        return self.content

    def __repr__(self) -> str:
        return self.content


class Script(SendableMixin):
    """
    Class for a generic, parameterisable script.

    Args:
        template (str):
            Base script to use. Accepts #parameters#
    """

    __slots__ = [
        "_template",
        "_subs",
        "_empty_treatment",
        "_init_args",
        "_temporary_args",
    ]

    def __init__(
        self,
        template: Union[str, None] = None,
        template_path: Union[str, None] = None,
        empty_treatment: str | None = None,
        **init_args,
    ):

        if template is None:
            if template_path is None:
                raise ValueError("Either template or template_path must be provided")
            with open(file=template_path, mode="r", encoding="utf-8") as o:
                template = o.read()

        self._template: str = self._parse_input(template)

        self._subs: dict[str, Substitution] = {}
        self._init_args = init_args
        self._temporary_args = {}

        self._empty_treatment = None
        self.empty_treatment = empty_treatment

        self._extract_subs()

    def __getattr__(self, item):
        if item != "_subs" and hasattr(self, "_subs") and item in self._subs:
            val = self._subs[item]
            logger.debug("returning alt __getattribute__ %s=%s", item, val)
            return val
        return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        try:
            self._subs[key].value = value
            if key in self._init_args:
                del self._init_args[key]
        except (AttributeError, KeyError):
            object.__setattr__(self, key, value)

    def __hash__(self) -> int:
        """
        We should return the hash of the template, not the output

        This means that two "different" scripts will be treated as equivalent,
        but if we're relying on the uuid of a script to generate a Dataset UUID for a
        non-function run, then we are physically unable to generate the UUID at init
        """
        return hash(self.template)

    def _parse_input(self, inp: Union[str, PathLike]) -> str:
        """
        Takes an input and checks if it is a file

        Args:
            inp (str, PathLike):
                input to check. May either be a "raw" template,
                or a path to a file containing one, for example
        """
        if os.path.isfile(inp) or isinstance(inp, PathLike):
            print(f"reading template from file: {inp}")
            with open(inp, mode="r", encoding="utf8") as o:
                data = o.read()
            return data
        return inp

    @property
    def uuid(self) -> str:
        """return the uuid of this script"""
        return generate_uuid(self.template)

    @property
    def short_uuid(self) -> str:
        """return the shortened uuid of this script"""
        return self.uuid[:8]

    def _extract_subs(self) -> None:
        """
        Extract all substitution objects from the template
        """
        # wipe cache
        self._subs = {}
        # extract all replacement targets
        symbols = re.findall(PLACEHOLDER_PATTERN, self.template)  # noqa: W605
        logger.info("Found substitution targets within script:%s", symbols)

        for match in symbols:
            logger.debug("processing match %s", match)
            name = match[0].lower()
            kwargs = match[1]

            if name in self._subs:
                if kwargs is not None and kwargs != "":
                    raise ValueError(
                        f"Got more kwargs for already registered argument "
                        f"{name}: {kwargs}"
                    )
                logger.debug("\talready processed, continuing")
                continue
            if kwargs == "":
                tmp = Substitution.from_string(name)
            else:
                tmp = Substitution.from_string(":".join(match))
            existing = getattr(self, tmp.name, None)
            if existing is not None and not isinstance(existing, DynamicMixin):
                raise ValueError(
                    f'Variable "{tmp.name}" already exists. This could '
                    f"cause unintended behaviour, please choose another "
                    f"name"
                )

            self._subs[tmp.name] = tmp

    def _link_subs(self) -> None:
        for sub in self.sub_objects:
            self._process_links(sub)

    def _process_links(self, sub: Substitution) -> None:
        if sub.static or sub.linked:
            return
        value_cache = sub.value

        logger.debug("linking sub %s, value=%s", sub.name, sub.value)

        # find all instances of an {expandable} section
        to_parse = _get_expandables(str(sub.value))

        if len(to_parse) == 0:
            return

        for section in to_parse:
            tokenized = Tokenizer(section)

            for sym in self.subs:
                if sym in tokenized.names:
                    tokenized.exchange_name(sym, f"self.{sym}")
                    if not self._subs[sym].linked:
                        self._process_links(self._subs[sym])
            logger.debug("evaluating exchanged source: %s", tokenized.source)
            try:
                evaluated = eval(tokenized.source)  # pylint: disable=eval-used
            except Exception as ex:
                raise ValueError(
                    f"Exception ({type(ex)}) when "
                    f"evaluating source in: {section}\n{ex}"
                ) from ex
            try:
                evaluated = evaluated.value
            except AttributeError:
                pass

            # if the whole value is to be evaluated, we can preserve the type
            preserve_type = (
                len(to_parse) <= 1
                and str(sub.value).startswith("{")
                and str(sub.value).endswith("}")
            )
            if preserve_type:
                if evaluated is None:
                    # None is a "protected" type, so passing it onward is dangerous
                    value_cache = "None"
                else:
                    value_cache = evaluated
            else:
                value_cache = value_cache.replace(f"{{{section}}}", str(evaluated))

        self._subs[sub.name].temporary_value = value_cache

    @property
    def template(self) -> str:
        """Returns the template"""
        return self._template

    @template.setter
    def template(self, template: str) -> None:
        """
        Update the template with a new one and regenerate the substitutions

        Args:
            template (str):
                new template to use
        """
        self._template = template
        self._extract_subs()

    @property
    def subs(self) -> list[str]:
        """Returns a list of all substitution names"""
        return list(self._subs.keys())

    @property
    def sub_objects(self) -> list[Substitution]:
        """Returns a list of all substitution objects"""
        return list(self._subs.values())

    @property
    def args(self) -> list:
        """Alias for self.subs"""
        return self.subs

    @property
    def arguments(self):
        """Alias for self.subs"""
        return self.subs

    @property
    def required(self) -> list[str]:
        """
        Returns a list of all required values

        Returns:
            list[str]: a list of all required values
        """
        required = []  # store the required values
        for sub in self.sub_objects:
            # if this sub is not optional, add it
            if not sub.optional:
                required.append(sub.name)
                required += sub.requires  # also add any dependencies
            # if this sub is replacing any others, check if *they* are required
            for name in sub.replaces:
                if not self._subs[name].optional:
                    required.append(sub.name)

            if not sub._optional:
                required += sub.requires

        return list(set(required))

    @property
    def missing(self) -> list[str]:
        """
        Returns a list of all missing required values

        Returns:
            list[str]: a list of all missing required values
        """
        output = []
        covered = []
        for sub in self.sub_objects:
            if not sub.optional and not sub.has_value:
                output.append(sub.name)

            for name in sub.requires:
                tmp = self._subs[name]
                if not tmp.has_value:
                    output.append(tmp.name)

            if sub.has_value:
                for name in sub.replaces:
                    covered.append(name)

        return list(set(output) - set(covered))

    @property
    def valid(self) -> bool:
        """
        Returns True if the script is currently "valid"

        Validation is defined as having no missing required values and all
        required values being present.

        ..note::
            Note that this property only considers values that have been flagged as
            required in the template. This essentially makes it up to the user to
            enable this functionality in their templates.

        Returns:
            bool: True if the script is currently "valid"
        """
        return len(self.missing) == 0

    @property
    def empty_treatment(self) -> Union[str, None]:
        """
        Returns the currently set global behaviour for empty treatment.
        """
        return self._empty_treatment

    @empty_treatment.setter
    def empty_treatment(self, style: Union[str, None]):
        if style is not None and style not in EMPTY_TREATMENT_STYLES:
            raise ValueError(INVALID_EMPTY_TREATMENT.format(style=style))
        self._empty_treatment = style

    def script(self, empty_treatment: Union[str, None] = None, **run_args) -> str:
        """
        Generate the script

        Args:
            empty_treatment (str, None):
                Overrides any local setting of ``empty_treatment`` if not None

        Returns:
            str: the formatted script
        """
        # check that empty_treatment is valid
        if empty_treatment is None:
            empty_treatment = self.empty_treatment
        if (
            empty_treatment is not None
            and empty_treatment not in EMPTY_TREATMENT_STYLES
        ):
            raise ValueError(INVALID_EMPTY_TREATMENT.format(style=empty_treatment))
        # update the values
        for k, v in self._init_args.items():
            if k in self.subs:
                self._subs[k].temporary_value = v
        for k, v in run_args.items():
            if k in self.subs:
                self._subs[k].temporary_value = v
        # check validity
        if not self.valid:
            raise ValueError(f"Missing values for parameters:\n{self.missing}")
        # generation section
        self._link_subs()  # ensure values are properly linked
        script = copy.deepcopy(self._template)  # do not clobber the internal template
        for sub in self.sub_objects:
            if sub.hidden:  # don't print any values that are hidden
                value = DELETION_FLAG_LINE
            else:
                value = sub.value  # get the value in string form

            if value is None or value == "None":
                # no value, triage this argument
                treatment = empty_treatment or sub.empty_treatment
                if treatment == "ignore":
                    continue
                elif treatment == "line":
                    value = DELETION_FLAG_LINE
                elif treatment == "local":
                    value = DELETION_FLAG_LOCAL
            # replace any instance of #name#, capturing args.
            # avoid any whitespace to not get tripped up by comments
            search = re.compile(
                REPLACEMENT_PATTERN.format(name=sub.name), re.IGNORECASE
            )
            script = re.sub(search, EscapeStub(value), script)
        # replacements complete, generate the output while handling missing values
        output = []
        for line in script.split("\n"):
            if DELETION_FLAG_LINE in line:
                continue
            output.append(line.replace(DELETION_FLAG_LOCAL, ""))
        # reset temporaries
        for sub in self.sub_objects:
            sub.reset_temporary_value()
        self._temporary_args = {}

        return re.sub(
            ESCAPE_SEQ_PATTERN, "", "\n".join(output)
        )  # remove any escape sequences

    def pack(self, values: bool = True, file: str | None = None) -> str:  # type: ignore
        """
        Store the Script

        Args:
            values (bool):
                includes any set values if True
            file (str):
                path to save to, returns the content if None

        Returns:
            str:
                File path if file is not None, else the storage content
        """
        output = copy.deepcopy(self.template)
        if values:
            for sub in self.sub_objects:
                search = re.compile(
                    REPLACEMENT_PATTERN.format(name=sub.name), re.IGNORECASE
                )
                if sub.value is not None and sub.value != sub.default:
                    value_insert = f"#{sub.target}:value={sub.value}#"
                    output = re.sub(search, EscapeStub(value_insert), output, count=1)

        if file is not None:
            with open(file, "w+", encoding="utf8") as o:
                o.write(output)
            return os.path.abspath(file)

        return output

    @classmethod
    def unpack(cls, input: str):  # type: ignore
        if os.path.exists(input):
            with open(input, "r", encoding="utf8") as o:
                input = o.read()

        return cls(template=input)


def _get_expandables(string: str) -> List[str]:
    output = []

    cache = []
    inset = []
    escape = False
    for char in string:
        if char == "\\" and not escape:
            escape = True
            continue

        if not escape:
            if char == "{":
                inset.append("{")

                # do not append the first instance
                if len(inset) == 1:
                    continue

            if char == "}":
                if len(inset) != 0:
                    if inset[-1] == "{":
                        del inset[-1]

                if len(inset) == 0:
                    output.append("".join(cache).strip())
                    cache = []
                    inset = []
                    continue
        else:
            escape = False

        if len(inset) == 0:
            continue

        cache.append(char)

    if len(cache) != 0:
        output.append("".join(cache).strip())

    return output



class DetailMixin:
    """
    Mixin class to add get_*_log methods to a Dataset.
    """

    _append_log = {}
    runners = []

    def get_append_log(self) -> str:
        """Returns the log from the previous append session"""
        output = []
        for session, log in self._append_log.items():
            if len(self._append_log) <= 1:
                output = [f"{line}" for line in log]
                break

            output.append(f"{session}:")

            output += [f"\t{line}" for line in log]

        return "\n".join(output)

    def get_run_log(self) -> str:
        """
        Retrieves the log of the previous run attempt, providing more details
        """
        output = []
        for runner in self.runners:
            if runner.state > "transferred":
                output.append(f"{runner}: {runner._run_state}")

        return "\n".join(output)

    def print_run_log(self) -> None:
        """Prints the log of the previous run attempt"""
        print(self.get_run_log())



from typing import Union, Any


def format_time(time: Union[int, float, str]) -> Union[str, None]:
    """
    Take integer seconds and generate a HH:MM:SS timestring

    Args:
        time (int):
            seconds
    Returns:
        (str):
            HH:MM:SS format timestamp
    """
    if time is None:
        return None
    if isinstance(time, str):
        # if given a string, first attempt to convert from semantic time
        if "d" in time or "h" in time or "m" in time or "s" in time:
            time = semantic_to_int(time)
        else:
            return time
    elif isinstance(time, float):
        time = int(time)

    return time_to_string(time)


def time_to_string(time: int) -> str:
    """Converts integer seconds to HH:MM:SS format"""
    mins = time // 60
    hours = mins // 60

    secstring = str(time % 60).rjust(2, "0")
    minstring = str(mins % 60).rjust(2, "0")
    hourstring = str(hours).rjust(2, "0")

    return f"{hourstring}:{minstring}:{secstring}"


def time_to_s(time: str) -> int:
    """Convert back from HH:MM:SS to integer seconds"""
    hh, mm, ss = time.split(":")

    return int(hh) * 3600 + int(mm) * 60 + int(ss)


def semantic_to_int(time: str) -> int:
    """
    Convert "semantic" time strings to integer format

    i.e. 24h => 86400, 30m => 1800
    """
    values = {"d": 86400, "h": 3600, "m": 60, "s": 1}

    sum_time = 0
    cache = []
    char = 0
    # iterate over string
    for char in time:
        try:  # store the numbers in the cache
            cache.append(int(char))
        except ValueError:
            # when we hit a non-integer, concat the cache and add it to the sum
            value = int("".join([str(c) for c in cache]))
            cache = []
            if char in values:
                sum_time += values[char] * value

    # convert to integer seconds
    return sum_time


def try_value(inp: Any) -> Any:
    """
    Try to access the value property

    This _needs_ to be used in the presence of DynamicValues, since they override
    a lot of basic functions. This can cause recursion errors if not handled properly
    """
    try:
        return inp.value
    except AttributeError:
        return inp



import importlib
import inspect
import logging
import os
import re
import typing
from typing import Union

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils.tokenizer import Tokenizer
from remotemanager.utils.uuid import generate_uuid

logger = logging.getLogger(__name__)


_SCRIPT_TEMPLATE = """
{function}

if __name__ == '__main__':
\tkwargs = {args}

\tresult = {name}(**kwargs)
"""

SIG_END_REGEX = re.compile(r"\).*:")

ARGS_REGEX_PATTERN = r"(\s|\()(\*{1}[\w]+)"
KWARGS_REGEX_PATTERN = r"(\s|\()(\*{2}[\w]+)"


def line_is_end_of_function(line: str) -> bool:
    """Returns True if the line is the end of a function definition"""
    return re.search(SIG_END_REGEX, line) is not None


class Function(SendableMixin):
    """
    Serialise function to an executable python file for transfer

    Args:
        func:
            python function for serialisation
    """

    def __init__(self, func: [typing.Callable, str], force_self: bool = False):
        logger.debug("creating new serialisable function for %s", func)

        self._uuid = ""
        # collect the source, name and signature
        if isinstance(func, str):
            source = func
            # need to extract the name and signature
            name = re.search(r"def\s(\w+)\(", source).group(1)

            signature = []
            for line in func.split("\n"):
                line = line.split("#", maxsplit=1)[0]
                if line.startswith("def"):
                    # drop the "def funcname(", keep the signature
                    line = line.split("(", maxsplit=1)[1]

                signature.append(line)
                if line_is_end_of_function(line):
                    # if we've reached the end of the signature, break
                    break

            signature = "\n".join(signature).strip(":")

        else:
            source = inspect.getsource(func)
            name = func.__name__
            signature = inspect.signature(func)

        if name in ["remote_load", "remote_dump"]:
            raise ValueError(f'function name "{name}" is protected')

        tmp = []
        indent = -1
        is_self = force_self  # default to False, allow forcing
        append = False  # don't begin the append until we hit the `def ...` line
        for line in source.split("\n"):
            if indent == -1 and line.strip().startswith("def"):
                if "(self" in line:
                    is_self = True
                append = True
                indent = line.index("def")

            if append:
                tmp.append(line[indent:])

        source = "\n".join(tmp).strip()

        signature = Function.prepare_signature(signature, is_self=is_self)

        self._signature = signature
        self._arglist = None
        self._fname = name

        logger.debug("updated signature to %s", self._signature)
        rawsig = Function.get_raw_signature(source)
        logger.debug("raw typed signature detected as %s", rawsig)

        source = self.replace_signature(source, rawsig, self.signature)

        if "**" not in source:  # check only for ** logic, since it could be renamed
            raise RuntimeError(f"**kwarg insertion failed for function:\n{source}")

        self._source = source

        self._uuid = generate_uuid(self._source)

    def __call__(self, *args, **kwargs):
        return self.object(*args, **kwargs)

    @staticmethod
    def get_raw_signature(source):
        """
        Strips the signature as it is typed.
        inspect.signature does some formatting which can cause replacement to
        break in some conditions

        Args:
            source (str):
                raw source

        Returns:
            (str): raw signature as typed
        """
        definition = []
        for line in source.split("\n"):
            line = line.split("#")[0].strip()
            # we need to split from the function name for the raw signature
            if "(" in line:
                # split from first ( and up,
                # then join for any tuples defined as default
                line = "(" + "(".join(line.split("(")[1:])

            # appending the final ":" causes issues, strip that in a safe
            # manner then append
            definition.append(line.rstrip(" ").rstrip(":"))

            # we've reached the end of the definition
            if line.endswith(":"):
                break

        return "\n".join(definition)

    @staticmethod
    def prepare_signature(sig, is_self: bool = False) -> str:
        """
        Inserts *args and **kwargs into any signature that does not already
        have it

        Args:
            sig:
                inspect.signature(func)
            is_self:
                inspect.signature ignores `self`, yet we want to preserve it.
                Adds the argument if True

        Returns:
            (str): formatted sig
        """
        if isinstance(sig, str):
            match = re.compile(r"->[^)]*:")
            return_annotation = re.findall(match, sig + ":")

            if len(return_annotation) == 0:
                return_annotation = None
            else:
                return_annotation = return_annotation[0].strip("->").strip(":").strip()

            sig = sig[: sig.rindex(")")]
            sig = sig.strip("(")
            args = [a.strip() for a in sig.split(",") if a.strip() != ""]
        else:
            args = [str(a) for a in sig.parameters.values()]

            return_annotation = sig.return_annotation

            # if return_annotation is None, we can't call "__name__" on it directly
            if return_annotation is not None:
                return_annotation = return_annotation.__name__
            else:
                return_annotation = "None"

        if return_annotation == "_empty":
            return_annotation = None

        if is_self and "self" not in args:
            args = ["self"] + args

        has_args = re.search(ARGS_REGEX_PATTERN, f"({sig})") is not None
        has_kwargs = re.search(KWARGS_REGEX_PATTERN, f"({sig})") is not None

        if not has_args:
            if has_kwargs:
                args.insert(-1, "*args")
            else:
                args.append("*args")

        if not has_kwargs:
            args.append("**kwargs")

        signature = f'({", ".join(args)})'
        if return_annotation is None:
            return signature
        return f"{signature} -> {return_annotation}"

    def replace_signature(self, source: str, rawsig: str, signature: str) -> str:
        """
        Performs replacement of the original "raw" signature
        with one with inserted *args, **kwargs
        """
        if len(rawsig.split("\n")) == 1:
            # direct replacement
            return source.replace(rawsig, signature, 1)
        # if we have multiple lines, we can recreate the "true" signature
        output = []
        append = True
        replaced = False
        for line in source.split("\n"):
            if "def" in line and not replaced:
                output.append(f"def {self.name}{self.signature}:")
                append = False
                replaced = True

            if append:
                output.append(line.rstrip())
            else:
                append = line_is_end_of_function(line)

        return "\n".join(output)

    @property
    def name(self) -> str:
        """
        Function name
        """
        return self._fname

    @property
    def raw_source(self) -> str:
        """
        Function source
        """
        return self._source

    @property
    def source(self) -> str:
        """
        Function source

        Returns:
            (str): source code
        """
        return self._source

    @property
    def return_annotation(self) -> Union[str, None]:
        if "->" in self.signature:
            return self.signature.split("->")[-1].strip()
        return None

    @property
    def signature(self) -> str:
        return self._signature

    @property
    def args(self) -> list:
        """returns a list of arguments, without annotations or defaults"""
        if self._arglist is not None:
            return self._arglist

        names = Tokenizer(self.signature).names

        args_name = re.search(ARGS_REGEX_PATTERN, self.signature).groups()[-1].strip()
        kwargs_name = (
            re.search(KWARGS_REGEX_PATTERN, self.signature).groups()[-1].strip()
        )

        # extract everything up to but not including *args, **kwargs and annotation
        names = names[: names.index(args_name.strip("*"))]

        # if we strip all spaces from the signature, then valid tokens must
        # present in the form ",{token}" or "({token}"
        compact = self.signature.replace(" ", "")
        output = []
        for token in names:
            if f",{token}" in compact or f"({token}" in compact:
                output.append(token)
        output += [args_name, kwargs_name]
        self._arglist = output  # cache the args
        return output

    @property
    def uuid(self) -> str:
        """
        Function uuid (64 characters)
        """
        return self._uuid

    @property
    def object(self):
        """
        Recreates the function object by writing out the source, and importing.

        Returns:
            typing.Callable:
                the originally passed function
        """

        tmp_file = ""
        try:
            tmp_file = os.path.abspath(f"{self.uuid}.py")

            with open(tmp_file, "w+") as o:
                o.write(self.source)

            spec = importlib.util.spec_from_file_location(self.uuid, tmp_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            func_object = getattr(module, f"{self.name}")

        finally:
            os.remove(tmp_file)

        return func_object

    def dump_to_string(self, args) -> str:
        """
        Dump this function to a serialised string, ready to be written to a
        python file

        Args:
            args (dict):
                arguments to be used for this dumped run

        Returns:
            (str):
                serialised file
        """

        if args is None:
            args = {}

        return _SCRIPT_TEMPLATE.format(
            **{"function": self.source, "name": self.name, "args": args}
        )



"""
DynamicValue stub class allows for deferred calculation of values.

Constructing a "tree" of values using these objects allows for later
assessment. Used in Computers for dynamic resource assignment.

>>> val_a = DynamicValue(10)
>>> val_b = DynamicValue(6)
>>> val_c = DynamicValue(val_a + val_b)
>>> val_c.value
16
"""

import logging
import math
from numbers import Number
from typing import Union, Any

from remotemanager.connection.computers.utils import format_time, try_value
from remotemanager.utils import ensure_list

logger = logging.getLogger(__name__)

EMPTY_TREATMENT_STYLES = [
    "line",  # default, removes the whole line
    "local",  # locally removes the value, ignoring the line
    "ignore",  # do nothing, performing no replacement
]

INVALID_EMPTY_TREATMENT = "empty_treatment must be one of the available styles: {style}"


class ChainingMixin:
    """Adds chaining ability for DynamicMixin and DynamicValue"""

    __slots__ = []

    @property
    def shortform_op(self) -> Union[str, None]:
        """
        Returns the operator in a readable form for calc insertion

        eg +, -, * instead of add, sub, mul, etc.
        """
        op = getattr(self, "op", None)
        ops = {
            "pow": "**",
            "div": "/",
            "mul": "*",
            "add": "+",
            "sub": "-",
            "eq": "==",
            "ne": "!=",
            "le": "<=",
            "lt": "<",
            "ge": ">=",
            "gt": ">",
            "get": "get",
        }
        return ops.get(op, None)

    def __pow__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "pow")

    def __mul__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "mul")

    def __truediv__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "div")

    def __add__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "add")

    def __sub__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "sub")

    def __eq__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "eq")

    def __ne__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "ne")

    def __le__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "le")

    def __lt__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "lt")

    def __ge__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "ge")

    def __gt__(self, other: Any) -> "DynamicValue":  # noqa: F821
        return self._chain_operation(other, "gt")

    def __getitem__(self, item):
        return self._chain_operation(item, "get")

    def __bool__(self):
        return bool(self.value)

    def _chain_operation(self, other, operation) -> NotImplemented:
        return NotImplemented

    @property
    def value(self) -> NotImplemented:
        return NotImplemented


class DynamicMixin(ChainingMixin):
    """
    Provides functions to enable Entities using DynamicValue to chain properly

    .. important::
        The DynamicValue in question must be directly available at `_value`

    Args:
        assignment (str):
            The variable to which this object is assigned, for introspection
        default (Any, None):
            Default value, marks this Resource as optional if present
        value (Any, None):
            Sets the value directly. You should ideally set the default, as this is easy
            to override and break
        optional (bool):
            Marks this resource as Optional. Required as there are actually
            three states:
                - Required input, required by scheduler.
                - Optional input, required by scheduler.
                - Optional input, optional by scheduler.
        requires (str, list):
            Stores the name(s) of another variable which is required alongside this one
        replaces (str, list):
            Stores the name(s) of another variable which is replaced by this one
        hidden (bool):
            The value will be present, but requests to not be displayed
        min (int):
            Minimum value for numeric inputs
        max (int):
            Maximum value for numeric inputs
        format (str):
            Expected format for number. Allows None, "time" or "float"
        empty_treatment (bool):
            Dictates how an empty value is treated. Possible options are:
            - line
                default, removes the whole line
            - local
                locally removes the output
            - ignore
                does nothing, leaves the output unchanged
        static (bool):
            Does not chain with other values if True
    """

    __slots__ = [
        "_name",
        "_optional",
        "_requires",
        "_replaces",
        "hidden",
        "_min",
        "_max",
        "_value",
        "format",
        "_empty_treatment",
        "static",
        "processed",
    ]

    def __init__(
        self,
        assignment: Union[str, None] = None,
        default: Union[Any, None] = None,
        value: Union[Any, None] = None,
        optional: bool = True,
        requires: Union[str, list, None] = None,
        replaces: Union[str, list, None] = None,
        hidden: bool = False,
        min: Union[int, None] = None,
        max: Union[int, None] = None,
        format: Union[str, None] = None,
        empty_treatment: str = "line",
        static: bool = False,
    ):
        self._name = assignment

        if optional in ["False", "false"]:
            optional = False
        elif optional == ["True", "true"]:
            optional = True
        self._optional = optional

        self._requires = ensure_list(requires, semantic=True)
        self._replaces = ensure_list(replaces, semantic=True)

        self.hidden = hidden

        self._min = min
        self._max = max

        self.format = format

        self._empty_treatment = None
        self.empty_treatment = empty_treatment

        self.static = static

        if isinstance(default, DynamicMixin):
            # assigning a default directly to a Resource or Substitution object causes
            # that object to be added, instead of a DynamicValue
            default = default._value

        if isinstance(value, DynamicMixin):
            value = value._value

        default = entry_format(default, format)
        value = entry_format(value, format)

        self._value = DynamicValue(
            a=value, b=None, op=None, default=default, assignment=assignment
        )

    def _chain_operation(self, other: Any, operation: str) -> "DynamicValue":
        try:
            other = other._value
        except AttributeError:
            pass
        obj = DynamicValue(self._value, other, operation)
        return obj

    @property
    def name(self) -> str:
        """Returns the name under which this resource is stored"""
        return self._name

    @property
    def min(self) -> Union[int, None]:
        """Minimal numeric value"""
        if self._min is not None:
            return int(self._min)
        return self._min

    @property
    def max(self) -> Union[int, None]:
        """Maximal numeric value"""
        if self._max is not None:
            return int(self._max)
        return self._max

    @property
    def default(self):
        """Returns the default, if available"""
        return self._format_value(self._value.default)

    @default.setter
    def default(self, default):
        self._value.default = default

    @property
    def _default(self):
        """Returns the "raw" default"""
        return self._value.default

    @property
    def optional(self):
        """Returns True if this Resource is optional at Dataset level"""
        return self._value.default is not None or self._optional

    @property
    def replaces(self) -> list:
        """
        List of arguments whom are no longer considered `required` if this
        resource is specified
        """
        return self._replaces

    @property
    def requires(self) -> list:
        """
        List of requirements if this resource is specified.
        e.g. nodes for mpi_per_node
        """
        return self._requires

    @property
    def empty_treatment(self):
        """
        Dictates how this parameter is treated if it is empty.

        Possible behaviours:

        ``line``:
            This is the default, and deletes the whole line. Useful when
            parameterising resource requests such as

            ``#SBATCH nodes=#nodes#``

        ``local``:
            Locally deletes the value, leaving the remainder of the line.
            Note that this does not delete anything _outside_ of the extents of the
            parameter. For example:

            ``#SBATCH nodes=#NODES:empty_treatment=local#`` -> ``#SBATCH nodes=``

            Useful when arguments may follow one another:

            ``#a:empty_treatment=local# #b:empty_treatment=local#``

            In this case, a missing ``a`` or ``b`` will not delete the other
        ``ignore``:
            Disables any treatment, values will be left as is:

            ``#a=#a:empty_treatment=local#`` -> ``#a=#a:empty_treatment=local#``
        """
        return self._empty_treatment

    @empty_treatment.setter
    def empty_treatment(self, style: str):
        if style.lower() not in EMPTY_TREATMENT_STYLES:
            raise ValueError(INVALID_EMPTY_TREATMENT.format(style=style))
        self._empty_treatment = style.lower()

    def _format_value(self, val):
        try:
            val = val.value
        except AttributeError:
            pass

        if val is None:
            return None
        if isinstance(val, bool):
            return val

        try:
            if self.format == "float":
                return float(val)
            if self.format == "time":
                return format_time(val)
        except ValueError:
            pass

        # no hard formatting, apply default ceil()
        if isinstance(val, (int, float)):
            val = int(math.ceil(float(val)))

        return val

    @property
    def value(self):
        """Attempt to safely return the value (default) from self"""
        if self.default is not None and self._value is None:
            val = self.default
        else:
            val = self._value

        try:
            val = val.value
        except AttributeError:
            pass

        val = self._format_value(val)

        self._validate(val)

        return val

    @value.setter
    def value(self, value):
        self.set_value(value)

    @property
    def temporary_value(self):
        return self._value.temporary_value

    @temporary_value.setter
    def temporary_value(self, value):
        self._value.temporary_value = value

    def reset_temporary_value(self):
        """Reset the temporary value back to None"""
        self.temporary_value = None

    @property
    def linked(self) -> bool:
        """Returns True if this Value has been linked by Script"""
        return self.temporary_value is not None

    def set_value(self, value):
        """
        Sets the value, separating out the function allows for property overloading

        Since this function handles value setting for both Resource/Substitution
        AND the DynamicValues within, we have some extra edge cases to catch

        case 1
            We have a resource, and are setting the value
            to a static int
        case 2
            We have a resource and are setting the value
            to directly mirror another resource
        case 3
            We have a resource and are setting the value
            to be a combination of other resources (DV)
        """
        self._validate(value)

        if isinstance(self._value, DynamicValue):
            # we're setting on an Argument _value
            if isinstance(value, DynamicValue):
                # if _value has any extra properties,
                # need to be careful not to drop them
                self._value._a = value._a
                self._value._b = value._b
                self._value._op = value._op
            else:
                self._value.value = value
            return

        if isinstance(value, DynamicValue):
            self._value = value
        else:
            self._value = DynamicValue(value)

    @property
    def has_value(self) -> bool:
        return self.value is not None

    def _validate(self, value) -> None:
        """
        Validates `value`. Raises exception if invalid
        """
        try:
            value = value.value
        except AttributeError:
            pass

        try:
            value / 1
            isnumeric = True
        except TypeError:
            isnumeric = False

        if isnumeric:
            name = getattr(self, "name", None)
            nameinsert = ""
            if name is not None:
                nameinsert = f" for {name}"
            if self.min is not None and value < self.min:
                raise ValueError(
                    f"{value}{nameinsert} is less than minimum value {self.min}"
                )
            if self.max is not None and value > self.max:
                raise ValueError(
                    f"{value}{nameinsert} is more than maximum value {self.max}"
                )

    @property
    def reduced(self):
        return self._value.reduced

    def pack(self, collect_value: bool = True):
        """
        Packs this Dynamic object down to a dictionary for storage

        Args:
            collect_value:
                Also collects the stored value if True
        """
        self.reset_temporary_value()
        data = {}

        def get_reduction(attr: str):
            """Attempts to get the reduced form of a value, returning None otherwise"""
            val = getattr(self, attr, None)
            if val is not None:
                try:
                    val = val.reduced
                    # remove redundant outer brackets
                    if val.startswith("(") and val.endswith(")"):
                        val = val[1:-1]
                except AttributeError:
                    # if the value is a string, it should be stored quoted
                    return treat_for_storage(val)

            return val

        default = get_reduction("_default")
        if default is not None:
            data["default"] = default

        if collect_value:
            value = self._value
            if isinstance(value, DynamicValue):
                if value.reduced == value.assignment:
                    # values can self-refer,
                    # returning the assignment instead of the value
                    value = treat_for_storage(value._a)
                else:
                    value = value.reduced
                    if value.startswith("(") and value.endswith(")"):
                        value = value[1:-1]
            if try_value(value) is not None and try_value(value) != try_value(default):
                data["value"] = value

        if not self.optional:
            data["optional"] = False
        if len(self.requires) != 0:
            data["requires"] = self.requires
        if len(self.replaces) != 0:
            data["replaces"] = self.replaces
        if self.min is not None:
            data["min"] = self.min
        if self.max is not None:
            data["max"] = self.max
        if self.format is not None:
            data["format"] = self.format

        return data


class DynamicValue(ChainingMixin):
    """
    Args:
        a:
            "First" number in operation
        b:
            "Second" number in operation. Can be None,
            in which case this value is considered "toplevel"
        op:
            Operation to use. Can be None for toplevel values
        default:
            Default value can be set in case the primary value is
            set to None
    """

    __slots__ = ["_a", "_b", "_op", "_default", "_assignment", "temporary_value"]

    def __init__(
        self,
        a: Union[Number, "DynamicValue", None],
        b: Union[Number, "DynamicValue", None] = None,
        op: Union[str, None] = None,
        default: Union[Number, "DynamicValue", None] = None,
        assignment: Union[str, None] = None,
    ):
        if try_value(a) == "":
            a = None
        if try_value(b) == "":
            b = None
        if op == "":
            op = None
        if b is None and op is not None:
            raise ValueError("Operator specified without 2nd value")
        if b is not None and op is None:
            raise ValueError("Cannot specify 2nd value without operator")

        self._a = a
        self._b = b
        self._op = op
        self._default = default
        self.temporary_value = None

        self._assignment = assignment

    def _chain_operation(self, other: Union[Any], operation: str):
        obj = DynamicValue(self, other, operation)
        return obj

    def __repr__(self):
        op = self.shortform_op
        if op is None:
            return str(self._a)
        return f"DynamicValue({self._a}{op}{self._b})"

    @property
    def a(self):
        """
        Returns:
            Value of "first" number
        """
        try:
            val = self._a.value
        except (ValueError, AttributeError):
            val = self._a
        return val

    @property
    def b(self):
        """
        Returns:
            Value of "second" number
        """
        try:
            val = self._b.value
        except (ValueError, AttributeError):
            val = self._b
        return val

    @property
    def op(self):
        """
        Returns:
            Operation string
        """
        return self._op

    @property
    def default(self):
        """
        Returns:
            The default value
        """
        return self._default

    @default.setter
    def default(self, default):
        """default setter"""
        self._default = default

    @property
    def assignment(self) -> Union[str, None]:
        """The variable at which this value is assigned, if available"""
        return self._assignment

    @property
    def static(self) -> bool:
        """Returns True if this Dynamic variable is static, rather than dynamic"""
        return self.op is None and self.a is not None

    @property
    def value(self):
        """
        Calculates value by calling the whole chain of numbers

        Returns:
            Value
        """
        logger.debug("evaluating %s", self.assignment)
        if self.temporary_value is not None:
            val = self.temporary_value
            logger.debug("\treturning temporary value: %s", val)
            return val

        if self.static:
            val = self.a
            logger.debug("\tstatic, returning self.a %s", val)
            return val

        if self.b is None:
            try:
                default = self.default.value
            except AttributeError:
                default = self.default
            logger.debug("\tself.b is None, returning default: %s", default)
            return default
        a = self.a
        b = self.b

        logger.debug("\tproceeding with a=%s, b=%s, op=%s", a, b, self.op)

        if isinstance(a, str) or isinstance(b, str):
            if self.op == "add":
                a = str(a)
                b = str(b)

        try:
            if self.op == "eq":
                return a == b
            if self.op == "ne":
                return a != b
            if self.op == "ge":
                return a >= b
            if self.op == "gt":
                return a > b
            if self.op == "le":
                return a <= b
            if self.op == "lt":
                return a < b
            if self.op == "add":
                return a + b
            if self.op == "sub":
                return a - b
            if self.op == "div":
                return a / b
            if self.op == "mul":
                return a * b
            if self.op == "pow":
                return a**b
            if self.op == "get":
                return a[b]
        except TypeError:
            return None

    @value.setter
    def value(self, val):
        """
        It is possible to update the value of a toplevel DynamicValue

        Args:
            val:
                New Value
        """
        if self._b is not None:
            name = self.assignment or "val"
            print(f"WARNING! Dynamic chain broken when assigning {name}={val}")
            self._b = None
            self._op = None
        self._a = val

    @property
    def reduced(self) -> str:
        """
        Returns the string form "reduced" version of this DynamicValue.

        In theory this should be storable as text within a database, without losing
        dependency information

        Extracts details from the assignment, if provided. Else returns the value.

        e.g.
        .. code:: python
            a = Resource(name="a")
            b = Resource(name="b")
            c = Resource(name="c")

            c = a + b

            c.reduce
            > "(a + b)"
        """

        def chain_reduction(obj):
            """Tries to "chain" the reduction to any other reducible objects"""
            try:
                return obj.reduced
            except AttributeError:
                # if we're at the end of the chain,
                # preferably return the assignment
                return get_assignment(obj)

        def get_assignment(obj):
            """
            Preferably returns the assignment property,
            otherwise returning the string form of obj
            """
            assign = getattr(obj, "assignment", None)
            if assign is None:
                return treat_for_storage(obj)
            return assign

        op = self.shortform_op

        if op is None:
            return get_assignment(self)

        a = chain_reduction(self._a)
        if self._b is None:
            return a

        b = chain_reduction(self._b)
        return f"({a} {op} {b})"


def treat_for_storage(var: Any) -> str:
    """
    Ensures that var is stored properly

    Strings must be quoted, but _not_ those that are
    actually ints
    """
    if var is None:
        return None
    if isinstance(var, str):
        try:
            tmp = str(int(var))
            # make sure we don't cast any floats to int
            if "." in tmp:
                return str(float(var))
            return tmp
        except (TypeError, ValueError):
            return f'"{var}"'
    return var


def concat_basic(
    a: Union[int, "DynamicValue", "Resource", "Substitution"],  # noqa: F821
    b: Union[int, "DynamicValue", "Resource", "Substitution"],  # noqa: F821
) -> "DynamicValue":
    """
    Concat two values a and b

    Required in the case where we want to add a value to a string

    Since str + Resource will call the __add__ method of the str object,
    we get a TypeError

    We need to sidestep this by adding the string to the _value_,
    then reversing the value.

    Doing it this way calls the corrent DynamicMixin.__add__(...) function,
    rather than str.__add__(...)

    ..note::
        This function will at least _try_ to return a+b beforehand

    Args:
        a: first value
        b: second value

    Returns:
        a+b
    """
    logger.debug(
        "Attempting to concat a=%s (type %s), b=%s (type %s)", a, type(a), b, type(b)
    )
    if try_value(a) == "":
        return b
    if try_value(b) == "":
        return a

    try:
        # works only if a is a DynamicMixin instance
        return a + b
    except TypeError:
        # create the sum in reverse then swap _a and _b
        tmp = b + a

        b_store = tmp._b
        tmp._b = tmp._a
        tmp._a = b_store

        return tmp


def entry_format(val, format: Union[str, None] = None):
    if val is None:
        return None

    if isinstance(val, DynamicValue):
        return val

    if isinstance(val, bool):
        return val

    if format == "time":
        try:
            return int(val)
        except (TypeError, ValueError):
            return val.strip('"')

    try:
        val = float(val)

        if val % 1 == 0:
            return int(val)
        return val

    except ValueError:  # non numeric
        return val



"""
The Dependency module handles the chaining of Datasets.

It provides overrides for any linked Dataset to propagate required calls.
"""

import logging
import os.path
import warnings
from typing import List, Tuple, Union, TYPE_CHECKING

from remotemanager.logging_utils.verbosity import Verbosity
from remotemanager.storage.sendablemixin import SendableMixin

# TYPE_CHECKING is false at runtime, so does not cause a circular dependency
if TYPE_CHECKING:
    from remotemanager.dataset.dataset import Dataset

logger = logging.getLogger(__name__)


class Dependency(SendableMixin):
    """
    Main Dependency class.

    .. note::
        This class is internal and should not be interacted with directly.

    Handles the chaining of Datasets and stores the "network".
    It provides overrides for any linked Dataset to propagate required calls.
    """

    _do_not_package = ["_network", "_ds_list_cache"]

    _ds_list_cache = []

    def __init__(self):
        logger.info("new Dependency created")

        self._network: List[Tuple[str, str]] = []
        self._parents: List[str] = []
        self._children: List[str] = []

    def add_edge(self, primary: "Dataset", secondary: "Dataset") -> None:
        """
        Adds an "edge" to the dependency network.

        This edge represents a parent-child link.

        Args:
            primary (Dataset): Parent Dataset object.
            secondary (Dataset): Child Dataset object.
        """
        self._ds_list_cache = None  # invalidate cache
        pair = (primary, secondary)
        if pair not in self._network:
            logger.info("adding new edge %s", pair)

            self._parents.append(primary.short_uuid)
            self._children.append(secondary.short_uuid)

            self._network.append(pair)

    @property
    def network(self) -> List[Tuple["Dataset", "Dataset"]]:
        """Returns the internal dependency network.

        This is a list of tuples representing the parent-child relationship.

        Returns:
            List(Tuple(Dataset, Dataset)
        """
        return self._network

    def get_children(self, dataset: "Dataset") -> List["Dataset"]:
        """
        Collect and return all children for a given Dataset object.

        Args:
            dataset (Dataset): Dataset to query.

        Returns:
            List: List of children of the given Dataset.
        """
        tmp = [
            self.network[i][1]  # 1 is the parent
            for i in range(len(self._parents))
            if self._parents[i] == dataset.short_uuid
        ]
        return tmp

    def get_parents(self, dataset: "Dataset") -> List["Dataset"]:
        """
        Collect and return all parents for a given Dataset object.

        Args:
            dataset (Dataset): Dataset to query.

        Returns:
            List: List of parents of the given Dataset.
        """
        tmp = [
            self.network[i][0]  # 0 is the child
            for i in range(len(self._children))
            if self._children[i] == dataset.short_uuid
        ]
        return tmp

    @property
    def ds_list(self) -> List["Dataset"]:
        """
        Returns a flattened list of all datasets within the network.

        Returns:
            _type_: _description_
        """
        if getattr(self, "_ds_list_cache", None) is not None:
            return self._ds_list_cache

        datasets = []
        for pair in self.network:
            for ds in pair:
                if ds not in datasets:
                    datasets.append(ds)

        self._ds_list_cache = datasets

        return datasets

    @property
    def runners(self) -> List:
        """
        Returns a list of all runners within the network.
        """
        runners = []
        for pair in self.network:
            for ds in pair:
                for runner in ds.runners:
                    runners.append(runner)
        return runners

    def update_db(self):
        """
        Ensure all datasets within a dependency update their databases together
        """
        for ds in self.ds_list:
            ds.update_db(dependency_call=True)

    def remove_run(self, ident: Union[int, str, dict]) -> bool:
        """
        Chains the Dataset.remove_run function to all Datasets within the dependency.

        Args:
            ident (Union[int, str, dict]): Runner identifier.

        Returns:
            bool: True if all runners were removed successfully, False otherwise.
        """
        out = []
        for ds in self.ds_list:
            out.append(ds.remove_run(ident=ident, dependency_call=True))

        return all(out)

    def clear_runs(self) -> None:
        """
        Remove all Runners from the Datasets.
        """
        for ds in self.ds_list:
            ds.wipe_runs(dependency_call=True, confirm=False)

    def clear_results(self, wipe: bool) -> None:
        """
        Ensure all datasets within a dependency clear their results together

        Args:
            wipe (bool): Additionally delete the local resultfiles if True
        """
        for ds in self.ds_list:
            ds.reset_runs(wipe, dependency_call=True, confirm=False)

    def wipe_local(self, files_only: bool = False) -> None:
        """
        Clears out the local directory.

        Args:
            files_only (bool, optional): Attepts to only delete files.
            Defaults to False, deleting the directory.
        """
        for ds in self.ds_list:
            ds.wipe_local(files_only=files_only, dependency_call=True, confirm=False)

    def wipe_remote(self, files_only: bool = False) -> None:
        """
        Clears out the remote directory.

        Args:
            files_only (bool, optional): Attepts to only delete files.
            Defaults to False, deleting the directory.
        """
        for ds in self.ds_list:
            ds.wipe_remote(files_only=files_only, dependency_call=True, confirm=False)

    def hard_reset(self, files_only: bool = False) -> None:
        """
        Hard resets all datasets. This will wipe local and remote directories.

        Args:
            files_only (bool, optional): Attepts to only delete files.
            Defaults to False, deleting the directory.
        """
        for ds in self.ds_list:
            ds.hard_reset(files_only=files_only, dependency_call=True, confirm=False)

    def append_run(
        self, caller, chain_run_args, run_args, force, lazy, *args, **kwargs
    ):
        """
        Appends runs with the same args to all datasets

        Args:
            lazy:
            caller:
                (Dataset): The dataset which defers to the dependency
            chain_run_args (bool):
                for dependency runs, will not propagate run_args to other datasets in
                the chain if False (defaults True)
            run_args (dict):
                runner arguments
            force (bool):
                force append if True
            lazy (bool):
                do not update the database after this append (ensure you call
                ``update_db()`` after appends are complete, or use the
                ``lazy_append()`` contex)
            *args:
                append_run args
            **kwargs:
                append_run keyword args

        Returns:
            None
        """
        logger.info("appending run from %s", caller)

        datasets = self.ds_list
        logger.info("There are %s datasets in the chain)", len(datasets))

        if chain_run_args:
            logger.info("chain_args is True, propagating")
            kwargs.update(run_args)

        for ds in datasets:
            if ds == caller:
                caller_args = {k: v for k, v in kwargs.items()}
                caller_args.update(run_args)
                ds.append_run(
                    dependency_call=True, force=force, lazy=lazy, *args, **caller_args
                )
            else:
                ds.append_run(
                    dependency_call=True, force=force, lazy=lazy, *args, **kwargs
                )

            if not lazy:
                ds.database.update(ds.pack())

    def finish_append(self) -> None:
        """
        Propagates the completion of runner append to all datasets in the chain.
        """
        for ds in self.ds_list:
            ds.finish_append(dependency_call=True, print_summary=False)

    @staticmethod
    def get_runner_remote_filepath(runner, workdir: str, filetype: str) -> str:
        """
        Generates the relative remote path from workdir to a runner file
        The file is specified by filetype
        """
        file = None
        if filetype == "resultfile":
            file = runner.resultfile
        elif filetype == "runfile":
            file = runner.runfile

        if file is None:
            raise ValueError(f"unknown filetype {filetype}")

        file = file.relative_remote_path(workdir)
        if not os.path.isabs(file):
            file = os.path.join("$sourcedir", file)
        return file

    def stage(
        self,
        # uuids: List[str] = None,
        force: bool = False,
        extra: str = "",
        force_ignores_success: bool = False,
        verbose: Verbosity = None,
        **run_args,
    ) -> bool:
        """
        Handles the Dataset Staging for a dependency situation.
        Args:
            dry_run (bool):
                create files, but do not run
            extra: str
                extra text to add to runner jobscripts
            force_ignores_success (bool):
                If True, `force` will submit only failed runners
            verbose:
                Sets local verbose level
        """
        logger.info("dependency internal stage call")

        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.ds_list[0].verbose

        for ds in self.ds_list:
            if ds.remote_dir != self.ds_list[0].remote_dir:
                msg = (
                    "Chained datasets cannot have mismatched remote_dirs!: "
                    f"{ds.remote_dir} & {self.ds_list[0].remote_dir}\n"
                    "You can use separate run_dirs to provide runtime isolation"
                )
                logger.critical(msg)
                raise RuntimeError(msg)
            if len(ds.runners) != len(self.ds_list[0].runners):
                msg = (
                    "Datasets do not have matching numbers of runners!: "
                    f"{len(ds.runners)} vs {len(self.ds_list[0].runners)}"
                )
                logger.critical(msg)
                raise RuntimeError(msg)

        verbose.print("Staging Dependency", level=1)

        run_args["force"] = force

        if all(r.state == "staged" for r in self.runners) and not force:
            verbose.print("All runners already staged, skipping", level=1)
            return False

        # grab all global extra jobscript content from the datasets
        global_extra = []
        for ds in self.ds_list:
            if ds._global_run_extra is not None:
                global_extra.append(ds._global_run_extra)

        bash_cache = []
        for i, ds in enumerate(self.ds_list):
            staged = 0
            verbose.print(f"  [{i}] {ds}", level=1, end="... ")
            verbose.print("", level=2)  # newline for higher verbose levels

            ds_run_dir = ds.run_dir or ds.remote_dir
            if ds.submitter not in bash_cache:
                bash_cache.append(ds.submitter)

            parent_datasets = self.get_parents(ds)
            child_datasets = self.get_children(ds)
            if len(parent_datasets) > 1:
                warnings.warn(
                    "Multiple parents detected. "
                    "Variable passing in this instance is unstable!"
                )

            for i, runner in enumerate(ds.runners):
                # section checks that the parent result exists, exiting if not
                parent_check = []
                for parent in parent_datasets:
                    parent_resultfile = self.get_runner_remote_filepath(
                        parent.runners[i], ds_run_dir, "resultfile"
                    )
                    parent_runfile = self.get_runner_remote_filepath(
                        parent.runners[i], ds_run_dir, "runfile"
                    )

                    parent_check.append(
                        f'export parent_result="{parent_resultfile}"\n'
                        f"if [ ! -f $parent_result ]; then\n"
                        f'\techo "Parent result not found at '
                        f'$parent_result" >> "{runner.errorfile.name}" && exit 1;\n'
                        f"fi\n"
                    )

                    # TODO this is broken with multiple parents
                    lstr = (
                        f"runfile = os.path.expandvars('{parent_runfile}')\n"
                        f"resultfile = os.path.expandvars('{parent_resultfile}')\n"
                        f"if os.path.getmtime(runfile) > "
                        f"os.path.getmtime(resultfile):\n"
                        f'\traise RuntimeError("outdated '
                        f'result file for parent")\n'
                        f"repo.loaded = repo.{parent.serialiser.loadfunc_name}("
                        f"resultfile)"
                    )
                    runner._dependency_info["parent_import"] = lstr

                parent_check = "".join(parent_check)
                # section deals with submitting children
                child_submit = []
                for child in child_datasets:
                    child_runner = child.runners[i]
                    runline = child_runner.generate_runline(child=True)
                    child_submit.append(runline)

                ready = runner.stage(
                    python=ds.url.python,
                    repo=self.ds_list[0].repofile.name,
                    global_extra="\n".join(global_extra),
                    extra=extra,
                    parent_check=parent_check,
                    child_submit=child_submit,
                    force_ignores_success=force_ignores_success,
                    verbose=verbose,
                    **run_args,
                )

                ds.sanitise_run_arg_paths(runner.derived_run_args)

                if ready:
                    staged += 1
            verbose.print(f"Done, {staged}/{len(ds.runners)} Runners staged", level=1)

        if staged == 0:
            logger.info("no runners completed the run assessment, skipping")
            verbose.print("No Runners staged", level=1)
            return False

        # deal with master file directly
        master_content = [
            f"source {self.ds_list[0].bash_repo.name}\n### runs ###",
            "export sourcedir=$PWD",
        ]

        for runner in self.ds_list[0].runners:
            runline = runner.generate_runline(child=False)
            master_content.append(runline)

        self.ds_list[0].master_script.write(master_content)

        # generate python repository
        self.ds_list[0]._write_to_repo(skip_function=True)
        self.ds_list[0].repofile.append("\n### Functions ###")
        for ds in self.ds_list:
            content = [f"# function for {ds}:\n", ds.function.source]
            self.ds_list[0].repofile.append("".join(content))

        # generate bash repository
        self.ds_list[0]._write_to_bash_repo(bash_cache)

        verbose.print("Done", level=1)
        return True

    def transfer(
        self,
        # uuids: List[str] = None,
        force: bool = False,
        extra: str = "",
        force_ignores_success: bool = False,
        verbose: Verbosity = None,
        **run_args,
    ) -> bool:
        """
        Transfer the files to the remote
        """
        logger.info("dependency internal run call")

        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.ds_list[0].verbose

        staged = self.stage(
            dependency_call=True,
            extra=extra,
            force=force,
            force_ignores_success=force_ignores_success,
            verbose=verbose,
            **run_args,
        )
        # taken from Dataset.transfer
        uuid_cache = []
        for runner in self.runners:
            if not runner.derived_run_args.get("force", False):
                if runner.state < "staged":
                    continue
                if runner.state >= "transferred":
                    continue

            self.ds_list[0].transport.queue_for_push(runner.jobscript)
            if self.ds_list[0].function is not None:
                self.ds_list[0].transport.queue_for_push(runner.runfile)

            logger.info("queuing extra files to send")
            for file in runner.extra_files_send:
                self.ds_list[0].transport.queue_for_push(file)

            uuid_cache.append(runner.uuid)

        if len(uuid_cache) == 0 and not staged:
            verbose.print("No Transfer required", level=1)
            return False

        if len(uuid_cache) == 1:
            verbose.print("Transferring for 1 Runner", level=1)
        else:
            verbose.print(f"Transferring for {len(uuid_cache)} Runners", level=1)

        self.ds_list[0].prepare_for_transfer()

        # queue
        self.ds_list[0].transport.queue_for_push(self.ds_list[0].master_script)
        self.ds_list[0].transport.queue_for_push(self.ds_list[0].repofile)
        self.ds_list[0].transport.queue_for_push(self.ds_list[0].bash_repo)

        self.ds_list[0].transport.transfer()
        for ds in self.ds_list:
            ds.set_runner_states("transferred")
        return True

    def run(
        self,
        force: bool = False,
        dry_run: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
        # uuids: list = None,
        extra: str = "",
        force_ignores_success: bool = False,
        **run_args,
    ) -> bool:
        """
        Handles the Dataset Run for a dependency situation.

        Args:
            dry_run (bool):
                create files, but do not run
            extra: str
                extra text to add to runner jobscripts
            force_ignores_success (bool):
                If True, `force` will submit only failed runners
            verbose:
                Sets local verbose level
        """
        logger.info("dependency internal run call")

        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.ds_list[0].verbose

        run_args["force"] = force

        cmd = (
            f"cd {self.ds_list[0].remote_dir} && "
            f"{self.ds_list[0].url.shell} {self.ds_list[0].master_script.name}"
        )
        asynchronous = False
        if not dry_run:
            self.transfer(
                extra=extra,
                force_ignores_success=force_ignores_success,
                verbose=verbose,
                **run_args,
            )

            uuid_cache = []
            asynchronous = False
            # check first here, since that's what's actually being executed
            for runner in self.ds_list[0].runners:
                if not runner.derived_run_args.get("force", False):
                    if runner.state < "transferred":
                        continue
                    if runner.state >= "submit pending":
                        continue

                if runner.derived_run_args["asynchronous"]:
                    asynchronous = True

                uuid_cache.append(runner.uuid)

            if len(uuid_cache) == 0:
                verbose.print("Remotely executing 1 Runner", level=1)
                return False

            if len(uuid_cache) == 1:
                verbose.print("Remotely executing 1 Runner", level=1)
            else:
                verbose.print(f"Remotely executing {len(uuid_cache)} Runners", level=1)

            for ds in self.ds_list:
                ds.set_runner_states(state="submit pending")
        else:
            for ds in self.ds_list:
                ds.set_runner_states(state="dry run")

            msg = f"launch command: {cmd}"
            logger.info(msg)
            verbose.print(msg, 1)

        self.ds_list[0]._run_cmd = self.ds_list[0].url.cmd(
            cmd, asynchronous=asynchronous, dry_run=dry_run, verbose=verbose
        )
        return True

    def update_runners(self):
        """
        Manifest only needs to be collected once, then all the runners
        can be updated by that call
        """
        runners = []
        for ds in self.ds_list:
            runners += ds.runners

        self.ds_list[0].update_runners(runners=runners, dependency_call=True)

    def check_failure(self):
        """
        Raises a RuntimeError if an error is detected in any of the runners

        Relies on the runner.is_failed property
        """
        for ds in self.ds_list:
            for runner in ds.runners:
                if runner.is_failed:
                    ds.fetch_results()
                    raise RuntimeError(
                        f"Detected a failure in dataset {ds}:\n{ds.errors}"
                    )



import os.path

from remotemanager.connection.computers.base import BaseComputer
from remotemanager.connection.computers.computer import Computer
from remotemanager.connection.url import URL
from remotemanager.dataset.dataset import Dataset
from remotemanager.decorators.remotefunction import RemoteFunction
from remotemanager.decorators.sanzufunction import SanzuFunction
from remotemanager.logging_utils.log import Handler
from remotemanager.script.script import Script

__all__ = [
    "Dataset",
    "URL",
    "RemoteFunction",
    "Computer",
    "Script",
    "BaseComputer",
    "SanzuFunction",
]  # noqa: F405
__version__ = "0.13.10"

# attach a global Logger to the manager
Logger = Handler()  # noqa: F405


# ipython magic
def load_ipython_extension(ipython):
    from remotemanager.decorators.magic import RCell

    ipython.register_magics(RCell)


def get_package_root() -> str:
    """returns the abspath to the package root directory"""
    return os.path.normpath(
        os.path.join(os.path.abspath(__file__), os.pardir, os.pardir)
    )



import os
from typing import List

from remotemanager.storage import TrackedFile


class ExtraFilesMixin:
    """
    Provides extra files related functionality to Runner and Dataset

    Requires either:
    - implementation of _extra_files dict
    - implementation of extra_files_send/recv properties

    Also requires a run_path property
    """

    run_path = NotImplemented
    local_dir = NotImplemented
    _extra_files = {}

    def _convert_files(self, files: list, recv: bool = False) -> List[TrackedFile]:
        """
        Attempts to convert any
        """
        output = []
        for file in files:
            if isinstance(file, TrackedFile):
                output.append(file)
                continue

            if isinstance(file, str):
                # handle a string type path/to/file.ext structure
                path, name = os.path.split(file)

                rpath = self.run_path
                lpath = os.getcwd()

                if recv:
                    lpath = os.path.join(lpath, self.local_dir)
                    rpath = os.path.join(rpath, path)
                else:
                    lpath = os.path.join(lpath, path)

                output.append(TrackedFile(lpath, rpath, name))
                continue

            if isinstance(file, dict):
                # handle a dict type {path/to/file.ext: path} structure
                lpath, name = os.path.split(list(file.keys())[0])
                rpath = list(file.values())[0]

                if recv:
                    lpath = self.local_dir if lpath == "" else lpath
                    rpath = os.path.join(self.run_path, rpath)
                else:
                    lpath = os.path.join(os.getcwd(), lpath)
                    rpath = os.path.join(self.run_path, rpath)

                output.append(TrackedFile(lpath, rpath, name))
                continue

            raise ValueError(f"Unhandled filetype {type(file)}")

        return output

    @property
    def extra_files(self) -> dict:
        """
        Returns the extra files set for this runner
        """
        return {"send": self.extra_files_send, "recv": self.extra_files_recv}

    @property
    def extra_files_send(self) -> list:
        """Methods for Dataset, should be overridden in Runner"""
        return self._convert_files(self._extra_files["send"])

    @property
    def extra_files_recv(self) -> list:
        """Methods for Dataset, should be overridden in Runner"""
        return self._convert_files(self._extra_files["recv"], recv=True)



import copy
import importlib
import json
import logging
import os.path
import re
import typing
import warnings
from typing import Union

import yaml

try:
    from warnings import deprecated
except ImportError:
    from typing_extensions import deprecated

from remotemanager.connection.computers.dynamicvalue import (
    DynamicMixin,
    DynamicValue,
    concat_basic,
)
from remotemanager.connection.computers.resource import Resource, Resources
from remotemanager.connection.computers.substitution import Substitution
from remotemanager.connection.computers.utils import try_value
from remotemanager.connection.url import URL
from remotemanager.logging_utils.utils import format_iterable
from remotemanager.storage.function import Function
from remotemanager.storage.sendablemixin import get_class_storage, INTERNAL_STORAGE_KEYS
from remotemanager.utils.tokenizer import Tokenizer
from remotemanager.utils.uuid import generate_uuid
from remotemanager.utils.version import Version

logger = logging.getLogger(__name__)

DEPRECATION_WARNING = (
    "BaseComputer is deprecated, being replaced entirely by Computer"
    "\n(from remotemanager import Computer)"
)


@deprecated(DEPRECATION_WARNING)
class BaseComputer(URL):
    """
    Base computer module for HPC connection management.

    Extend this class for connecting to your machine
    """

    def __init__(self, template: Union[str, None] = None, **kwargs):
        super().__init__(**kwargs)

        self._template = None
        self.template = template

        self._temporary_args = {}

        self.resource_tag = "--"
        self.resource_separator = "="

        self._extra = ""
        self._internal_extra = ""

        self._super_issued = True  # enables check for super() call

    def __getattribute__(self, item):
        if (
            item != "_temporary_args"
            and hasattr(self, "_temporary_args")
            and item in self._temporary_args
        ):
            val = self._temporary_args[item]
            logger.debug("returning alt __getattribute__ %s=%s", item, val)
            return val
        return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        """
        If the set `key` attribute is a Resource, and `value` is not a Resource,
        instead set the `value` of that attribute

        Args:
            key:
                attribute name to set
            value:
                value to set to

        Returns:
            None
        """
        # if this resource already exists, and we're _not_ trying to set a new resource,
        # set the value of the Resource to value
        if key in self.__dict__ and isinstance(getattr(self, key), DynamicMixin):
            # need to get before we can set
            getattr(self, key).value = value
            logger.debug("used __setattr__ Resource override for %s=%s", key, value)
            return

        object.__setattr__(self, key, value)

    @property
    def uuid(self):
        seed = super().uuid
        base = format_iterable(self.arguments)

        return generate_uuid(seed + base)

    @property
    def short_uuid(self):
        return self.uuid[:8]

    def get_parser(self):
        parser = getattr(self, "parser", None)

        if parser is None:
            # legacy _parser
            parser = getattr(self, "_parser", None)

        return parser

    def _extract_subs(self, link: bool = True):
        symbols = re.finditer(r"#(\w+)(?::([^#]+))?#", self.template)  # noqa: W605
        logger.info("Found substitution targets within script:%s", symbols)
        processed = []
        for match in symbols:
            logger.debug("processing match %s", match)
            target = match.group(0)
            sym = match.group(1)
            kwargs = match.group(2)

            name = sym.lower()
            if name in processed:
                if kwargs is not None and kwargs != "":
                    raise ValueError(
                        f"Got more kwargs for already registered argument "
                        f"{name}: {target}"
                    )
                logger.debug("\talready processed, continuing")
                continue

            tmp = Substitution.from_string(target)

            existing = getattr(self, tmp.name, None)
            if existing is not None and not isinstance(existing, DynamicMixin):
                raise ValueError(
                    f'Variable "{tmp.name}" already exists. This could '
                    f"cause unintended behaviour, please choose another "
                    f"name"
                )

            setattr(self, tmp.name, tmp)

            processed.append(name)

        if link:
            self._link_dynamic_subs()

    def _link_dynamic_subs(self):
        # expand any {calculable} variables
        for sub in self.substitution_objects:
            default = sub.target_kwargs.get("default", None)
            # skip if None, not a string, or no {}
            if default is None or not isinstance(default, str) or "{" not in default:
                continue

            logger.debug("Further processing on default %s", default)

            main_cache = []
            proc_cache = []
            indent = 0
            for char in default:
                if char == "{":  # "indent" into brackets
                    indent += 1
                    continue
                if char == "}":  # "dedent" out of brackets
                    indent -= 1
                    # if we drop back into non processing stream, process what's in the
                    # process cache
                    if indent == 0:
                        tokenized = Tokenizer("".join(proc_cache))
                        for name in self.arguments:
                            tokenized.exchange_name(name, f"self.{name}")
                        logger.debug("\tevaluating source %s", tokenized.source)
                        try:
                            evaluated = eval(tokenized.source)
                        except Exception as ex:
                            raise RuntimeError(
                                f"{ex.__class__.__name__} when evaluating substitution "
                                f"{sub.name}: {tokenized.source}"
                                f"\n(see above)"
                            ) from ex
                        logger.debug("\tevaluated to: %s", evaluated)
                        main_cache.append(
                            evaluated
                        )  # This can be a DynamicValue, which needs to be handled later
                        proc_cache = []
                    continue

                if indent == 0:
                    # non processing stream, just append
                    main_cache.append(char)
                else:
                    proc_cache.append(char)

            logger.debug("Compacting cache %s", main_cache)
            # First, "squash" any strings
            # converts ["a", "b", "c"]
            # to ["abc"], respecting any breaks for dynamic values
            squashed = []
            tmp = []
            for char in main_cache:
                logger.debug(f"\tchar {char}")
                if isinstance(char, (DynamicValue, DynamicMixin, bool)) or char is None:
                    logger.debug(
                        "\t\tchar is dynamic or otherwise unjoinable, "
                        "add to output cache and continue"
                    )
                    squashed.append("".join(tmp))
                    squashed.append(char)
                    tmp = []
                else:
                    try:
                        char / 1
                        char = str(char)
                    except (TypeError, ValueError):
                        pass
                    logger.debug("\t\tadded to tmp, tmp is now: %s", tmp)
                    tmp.append(char)
            if len(tmp) != 0:
                logger.debug("\tfinal tmp comact: %s", tmp)
                squashed.append("".join(tmp))

            # then concat the items
            logger.debug("Pre concat list: %s", squashed)
            output = None
            for item in squashed:
                if try_value(item) == "" or item is None:
                    continue
                if output is None:
                    output = item
                else:
                    output = concat_basic(output, item)

            sub.default = output

    def unreduce_args(self) -> None:
        """
        Attempts to find any arguments who have had their chains "reduced", and
        unreduce them back into dynamic variables
        """

        def unreduce_attribute(tmp):
            replace = False
            if isinstance(tmp, str):
                if tmp.startswith('"'):
                    logger.debug(
                        """\tValue is quoted, stripping ", but not replacing"""
                    )
                    tmp = tmp.strip('"')
                elif tmp.startswith("'"):
                    logger.debug(
                        """\tValue is quoted, stripping ', but not replacing"""
                    )
                    tmp = tmp.strip("'")
                else:
                    replace = True

            if replace:
                replaced = False  # only evaluate if a replacement occurs
                for name in self.arguments:
                    if name in tmp:
                        tmp = tmp.replace(name, f"self.{name}")
                        replaced = True
                logger.debug("\tUpdated default for %s to %s.)", arg.name, tmp)
                if replaced:
                    try:
                        tmp = eval(tmp)
                        logger.debug("\tEvaluated to %s.)", tmp)
                    except Exception as E:
                        logger.warning(
                            "Got an exception when attempting to "
                            "evaluate the arg %s with default %s:\n%s",
                            arg.name,
                            tmp,
                            str(E),
                        )
                else:
                    logger.debug("\tNo replacement, no evaluation.")
                logger.debug(
                    "\tDone.)",
                )
            else:
                logger.debug(f"\tNot stringtype ({type(tmp)}, no replacement.")
            return tmp

        logger.debug("unreducing all args, %s", self.arguments)
        for arg in self.argument_objects:
            tmp = getattr(arg, "default", None)
            logger.debug("Treating default for %s=%s.)", arg.name, tmp)
            arg.default = unreduce_attribute(tmp)

    @classmethod
    def from_dict(cls, spec: dict, **url_args):
        """
        Create a Computer class from a `spec` dictionary. The required values are:

        - resources:
            a dict of required resources for the machine (mpi, nodes, queue, etc.)
        - resource_parser:
            a function which takes a dictionary of {resource: Option}, returning a list
            of valid jobscript lines

        You can also provide some optional arguments:

        - required_or:
            list of resources, ONE of which is required. Note that this must be a
            _list_ of dicts, one for each or "block"
        - optional_resources:
            as with `resources`, but these will not be stored as required values
        - optional_defaults:
            provide defaults for the names given in optional_resources. When adding the
            optional arg, the optional_defaults will be checked to see if a default is
            provided
        - host:
            machine hostname. Note that this _must_ be available for a job run, but if
            not provided within the spec, can be added later with
            ``url.host = 'hostname'``
        - submitter:
            override the default submitter
        - python:
            override the default python
        - extra:
            any extra lines that should be appended after the resource specification.
            Note that this includes module loads/swaps, but this can be specified on a
            per-job basis, rather than locking it into the `Computer`

        The `resources` specification is in `notebook`:`machine` order. That is to say
        that the `key` is what will be required in the _notebook_, and the `value` is
        what is placed in the jobscript:

        >>> spec = {'resources': {'mpi': 'ntasks'}, ...}
        >>> url = BaseComputer.from_dict(spec)
        >>> url.mpi = 12
        >>> url.script()
        >>> "--ntasks=12"

        Args:
            spec (dict):
                input dictionary
            url_args:
                any arguments to be passed directly to the created `url`

        Returns:
            Computer class as per input spec
        """
        from remotemanager import Logger

        payload = copy.deepcopy(spec)
        # initialise with the version checking
        version = payload.pop("remotemanager_version", None)
        if version is not None:
            version = Version(version)
            Logger.debug(f"unpacking file generated by remotemanager v{version}")
        else:
            Logger.debug("no version detected within file, proceeding with care")

        # check if we have a stored Computer Class
        class_store_key = INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"]
        if class_store_key in payload:
            class_storage = payload[class_store_key]
            Logger.debug(f"using class storage {class_storage}")
            # first, try grab the module
            try:
                mod = importlib.import_module(class_storage["mod"])
                # then try import from that module
                try:  # replace the cls if possible
                    cls = getattr(mod, class_storage["name"])
                except AttributeError:  # inner import error, no class
                    warnings.warn(f"Could not import class {class_storage['name']}")
            except ModuleNotFoundError:  # outer import error, no module
                warnings.warn(f"Could not import module {class_storage['mod']}")
        Logger.debug(f"from_dict called on {cls}")
        # now attempt to unpack the parser, if it exists
        parser = None
        parser_source = payload.pop("resource_parser_source", None)
        if parser_source is not None:
            Logger.debug("generating parser from source code")
            parser = unpack_parser(parser_source)
        else:  # legacy unpack method
            Logger.debug("parser source code not found at resource_parser_source")
            parser_source = payload.pop("resource_parser", None)
            parser = unpack_parser(parser_source)
        # parser object must be assigned (monkey patched) at class level, or it will
        # not be a bound-method, which is required for `self` and inspect to work
        if parser is not None:
            Logger.debug("assigning parser object at parser")
            cls.parser = parser.object
            cls._parser_source = parser.raw_source
        else:
            Logger.debug("parser NOT assigned")
        # create a new instance of the class
        computer = cls()
        # collect the resources and substitutions that are stored
        # these need to be handled later
        resources = payload.pop("resources")
        substitutions = payload.pop("substitutions", {})
        # `extra` is nuanced, needs to be split into internal and external
        external_extra = payload.pop("extra", "")
        setattr(computer, "_extra", external_extra)
        internal_extra = payload.pop("internal_extra", "")
        setattr(computer, "_internal_extra", internal_extra)

        # add bulk of the content first
        for key, val in payload.items():
            setattr(computer, key, val)
        for key, val in url_args.items():
            setattr(computer, key, val)

        # add the argument objects after
        # this allows any special settings to take priority
        oldstyle = False
        for field, resource_args in resources.items():
            if isinstance(resource_args, str):
                # this is a consequence of an older style "name": "flag" spec
                # cast to new type by setting it to "flag"
                # other args can be set later
                oldstyle = True
                resource_args = {"flag": resource_args, "optional": False}
            if "name" not in resource_args:
                resource_args["name"] = field
            resource = Resource(**resource_args)
            # we need to use the base object setattr, since we override it here
            object.__setattr__(computer, field, resource)

        for field, sub_args in substitutions.items():
            sub_args["name"] = field
            sub = Substitution(**sub_args)
            object.__setattr__(computer, field, sub)
        # legacy style spec has some extra fields to handle
        # update the computer inplace with a function
        if oldstyle:
            legacy_unpack(computer, payload)

        computer.unreduce_args()
        return computer

    @property
    def parser_source(self) -> str:
        if getattr(self, "_parser_source", None) is not None:
            parser = Function(self._parser_source, force_self=True).raw_source
        else:
            parser = self.get_parser()
            try:
                # avoids a strange error where the parser source cant be found.
                # likely loads it into memory where `inspect` can access it
                # noinspection PyStatementEffect
                parser
            except AttributeError:
                pass

        if isinstance(parser, Function):
            return parser.raw_source
        elif parser is not None:
            return Function(parser, force_self=True).raw_source

    def to_dict(
        self,
        include_extra: bool = True,
        include_version: bool = True,
        collect_values: bool = True,
    ) -> dict:
        """
        Generate a spec dict from this Computer

        Args:
            include_extra:
                includes the `extra` property if True (default True)
            include_extra:
                includes the current remotemanager version if True (default True)
            collect_value:
                Also collects the stored values of the arguments if True

        Returns:
            dict
        """
        from remotemanager import __version__

        logger.debug(f"to_dict called on {self}")

        # gather all non resource objects
        spec = {
            k: getattr(self, k)
            for k in self.__dict__
            if k not in self.arguments and not k.startswith("_")
        }

        spec["resources"] = {
            n: r.pack(collect_value=collect_values)
            for n, r in self.resource_dict.items()
        }

        spec["substitutions"] = {
            n: r.pack(collect_value=collect_values)
            for n, r in self.substitution_dict.items()
        }

        spec["resource_parser_source"] = self.parser_source

        # grab `extra` if requested
        spec["internal_extra"] = self._internal_extra
        if include_extra and self.extra is not None:
            spec["extra"] = try_value(self._extra)
        # round up missing package inclusions
        collect = [
            "submitter",
            "shebang",
            "python",
            "pragma",
            "host",
            "user",
            "port",
            "template",
        ]
        for name in collect:
            val = getattr(self, name, None)
            if val is None:
                continue
            spec[name] = val
        # parser object is not required, and looks ugly
        if "parser" in spec:
            del spec["parser"]
        cls_store_key = INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"]
        spec[cls_store_key] = get_class_storage(self)

        if include_version:
            spec["remotemanager_version"] = __version__

        return spec

    @classmethod
    def from_yaml(cls, filepath: str, **url_args):
        """
        Create a Computer from `filepath`.

        Args:
            filepath:
                path containing yaml computer spec
            **url_args:
                extra args to be passed to the internal URL

        Returns:
            BaseComputer
        """
        if isinstance(filepath, str):
            try:
                with open(filepath, "r") as o:
                    data = yaml.safe_load(o)
            except OSError:
                data = yaml.safe_load(filepath)
        else:
            data = yaml.safe_load(filepath)

        return cls.from_dict(data, **url_args)

    def to_yaml(
        self,
        filepath: Union[str, typing.IO, None] = None,
        include_extra: bool = True,
        include_version: bool = True,
        collect_values: bool = True,
    ) -> Union[str, None]:
        """
        Dump a computer to yaml `filepath`.

        Args:
            filepath:
                path containing yaml computer spec
            include_extra:
                includes the `extra` property if True (default True)
            collect_value:
                Also collects the stored values of the arguments if True
        """
        data = self.to_dict(
            include_extra=include_extra,
            include_version=include_version,
            collect_values=collect_values,
        )
        # source will simply not print correctly with base yaml
        # extract it and do it manually, if it exists
        if "resource_parser_source" in data:
            parser_string = ["resource_parser_source: |"] + [
                f"    {line}" for line in data.pop("resource_parser_source").split("\n")
            ]
            # dump the remaining content to string
            prepared = yaml.dump(data)
            # append the cleaned string
            prepared += "\n".join(parser_string)
        else:
            prepared = yaml.dump(data)

        if filepath is None:  # "dump" to string
            return prepared
        elif isinstance(filepath, str):  # dump to path
            with open(filepath, "w+") as o:
                o.write(prepared)
        else:  # assume file handler and dump there
            filepath.write(prepared)

    @staticmethod
    def download_file(file_url: str, filename: str) -> None:
        """
        Download file at url `file_url` and write the content out to `filename`

        Args:
            file_url: url of file
            filename: name to write content to
        """
        import requests

        response = requests.get(file_url)

        if response.status_code == requests.codes.ok:
            # Save the file
            fld, file = os.path.split(filename)
            if fld != "" and not os.path.exists(fld):
                os.makedirs(fld)

            with open(filename, "wb") as file:
                file.write(response.content)
            print(f"Grabbed file '{filename}'")
        else:
            raise RuntimeError(f"Could not find a file at: {file_url}")

    @classmethod
    def from_repo(
        cls,
        name: str,
        branch: str = "main",
        repo: str = "https://gitlab.com/l_sim/remotemanager-computers/",
        **url_args,
    ):
        """
        Attempt to access the remote-computers repo, and pull the computer with name
        `name`

        Args:
            name (str):
                computer name to target
            branch (str):
                repo branch (defaults to main)
            repo (str):
                repo web address (defaults to main l_sim repo)

        Returns:
            BaseComputer instance
        """
        from remotemanager.utils import ensure_filetype

        filename = ensure_filetype(name, "yaml").lower()
        url = f"{repo}-/raw/{branch}/storage/{filename}"

        print(f"polling url {url}")

        cls.download_file(url, filename)

        return cls.from_yaml(filename, **url_args)

    def generate_cell(
        self, name: Union[str, None] = None, return_string: bool = False
    ) -> Union[None, str]:
        """
        Prints out copyable source which regenerates this Computer

        Args:
            name (str, None):
                Optional name for new computer. Defaults to `new`
            return_string (bool):
                Also returns the string if True. Defaults to False

        Returns:
            (None, str)
        """
        if name is None:
            name = "new"
        output = [
            "# Copy the following into a jupyter cell or python script "
            "to generate a modifiable source",
            "\n# Parser source code",
        ]
        source = self.to_dict(include_version=False)

        try:
            output.append(self.parser_source)
            parser = Function(self.parser_source)

            source.pop("resource_parser_source")
            # use json.dumps with indent=4 to format dict
            output.append(
                f"\n# JSON compatibility\n"
                f"true = True\n"
                f"false = False\n\n"
                f"# spec dict\n\nspec = {json.dumps(source, indent=4)}"
            )
            output.append(f'spec["resource_parser"] = {parser.name}')

            output.append(f"\n{name} = BaseComputer.from_dict(spec)")

        except TypeError:
            output.append(
                f"# JSON compatibility\n"
                f"true = True\n"
                f"false = False\n\n"
                f"# spec dict\n\nspec = {json.dumps(source, indent=4)}"
            )
            output.append(f"\n{name} = BaseComputer.from_dict(spec)")

        output = "\n".join(output)

        print(output)

        if return_string:
            return output

    @property
    def is_super(self):
        if not hasattr(self, "_super_issued"):
            return False
        return self._super_issued

    @property
    def arguments(self) -> list:
        return sorted(
            [
                k
                for k, v in self.__dict__.items()
                if isinstance(v, (Resource, Substitution))
            ]
        )

    @property
    def substitutions(self) -> list:
        return sorted(
            [k for k, v in self.__dict__.items() if isinstance(v, Substitution)]
        )

    @property
    def resources(self) -> list:
        return sorted([k for k, v in self.__dict__.items() if isinstance(v, Resource)])

    @property
    def argument_objects(self) -> list:
        return [
            v
            for k, v in self.__dict__.items()
            if isinstance(v, (Resource, Substitution))
        ]

    @property
    def substitution_objects(self) -> list:
        return [v for k, v in self.__dict__.items() if isinstance(v, Substitution)]

    @property
    def resource_objects(self) -> list:
        return [v for k, v in self.__dict__.items() if isinstance(v, Resource)]

    @property
    def argument_dict(self) -> dict:
        return {k.name: k for k in self.argument_objects}

    @property
    def substitution_dict(self) -> dict:
        return {k.name: k for k in self.substitution_objects}

    @property
    def resource_dict(self) -> dict:
        return {k.name: k for k in self.resource_objects}

    @property
    def required(self) -> list:
        """
        Returns a list of required arguments
        """
        required = []

        def append_if(item):
            if item not in required:
                required.append(item)

        for name, resource in self.argument_dict.items():
            if not resource.optional:
                append_if(name)
            for name in resource.requires:
                append_if(name)

        return required

    @property
    def missing(self) -> list:
        """
        Returns the currently missing arguments
        """
        missing = []
        covered = []
        for resource in self.argument_objects:
            # this resource has a value, so is not missing, and can replace others
            if resource.value is not None:
                covered.append(resource.name)
                for name in resource.replaces:
                    if name in missing:
                        missing.remove(name)
                    covered.append(name)
            # resource is missing a value and is non optional
            elif not resource and not resource.optional:
                if resource.name not in covered:
                    missing.append(resource.name)

            for name in resource.requires:
                if name not in covered and not self.argument_dict[name]:
                    missing.append(name)

        return missing

    @property
    def valid(self) -> bool:
        """Returns True if there are no missing attributes"""
        return len(self.missing) == 0

    def parser(self, resources) -> list:
        """
        Default parser for use on basic "SLURM style" machines.

        Will iterate over resource objects, creating a script of the format:

        {pragma} --{flag}={value}

        ..note::
            This method can (and should) be overidden for a custom parser.

        Args:
            resources:
                Resources object, to be created by BaseComputer

        Returns:
            list of resource lines
        """
        output = []
        for r in resources:
            if r:
                output.append(r.resource_line)

        return output

    @property
    def extra(self):
        return try_value(self._internal_extra) + try_value(self._extra)

    @extra.setter
    def extra(self, external):
        self._extra = external

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, template):
        self._template = template
        if template is not None:
            self._extract_subs()

    def script(self, **kwargs) -> str:
        """
        Takes job arguments and produces a valid jobscript

        Args:
            insert_stub (add to kwargs):
                inserts the submission block stub if True (Used by Dataset)

        Returns:
            (str):
                script
        """
        if not self.is_super:
            raise RuntimeError(
                f"This Computer does not seem to have the correct attributes.\n"
                f"Is the super().__init__() line missing from this class' __init__?"
            )

        logger.info(f"script method called with run_args {kwargs}")
        # set temp args
        for k, v in kwargs.items():
            if k in self.arguments:
                self.argument_dict[k].temporary_value = v
            else:
                self._temporary_args[k] = v

        self._link_dynamic_subs()

        for key, val in kwargs.items():
            if key in self.arguments:
                logger.info("\tset temporary arg of %s to %s", key, val)
                self.argument_dict[key].temporary_value = val

        if not self.valid:
            raise RuntimeError(f"missing required arguments: {self.missing}")

        if self.template is None:
            logger.debug("Creating script from Resources")
            script = self._standard_script(**kwargs)
        else:
            logger.debug("Using template for generation")
            empty_treatment = kwargs.get("empty_treatment", "wipe")
            script = self.apply_substitutions(
                self.template.split("\n"), empty_treatment=empty_treatment
            )

        self._temporary_args = {}
        for arg in self.argument_objects:
            arg.reset_temporary_value()

        return script

    def _standard_script(self, **kwargs):
        pragma = getattr(self, "pragma", None)
        submit_args = Resources(
            resources=copy.deepcopy(self.resource_objects),
            pragma=pragma,
            tag=self.resource_tag,
            separator=self.resource_separator,
            run_args=kwargs,
        )

        script = [self.shebang]

        script += self.parser(submit_args)

        if self.extra is not None:
            script.append(self.extra)

        extras = ["global_extra", "runner_extra", "tmp_extra"]
        for key in extras:
            extra = kwargs.get(key, None)
            if extra is not None:
                script.append(extra)

        insert_stub = kwargs.pop("insert_stub", False)
        if insert_stub:
            script.append("#SUBMISSION_SUBSTITUTION#")

        script = self.apply_substitutions(script)

        return script

    def apply_substitutions(self, script: list, empty_treatment: str = "wipe") -> str:
        """
        Apply Substitution objects to the script

        Args:
            script: Base script to operate on
            empty_treatment: Defines the default behaviour for valueless args
                wipe: deletes the whole line (default)
                ignore: skip the value, leaving the substitution object there
                local: deletes only the missing argument
        """
        empty_behaviours = ["wipe", "ignore", "local"]

        if empty_treatment not in empty_behaviours:
            raise ValueError(
                f"empty_treatment={empty_treatment} is "
                f"not in available behaviours: {empty_behaviours}"
            )

        script = "\n".join(script)

        logger.debug("applying substitutions to script:\n%s", script)

        for sub in self.substitution_objects:
            logger.debug("processing Sub pointed at %s", sub.target)
            if sub.target == "":
                continue  # can't replace "" with something

            term = re.compile(f"#{sub.name}#", re.IGNORECASE)
            logger.debug("\tusing search term %s", term)

            value = sub.value
            logger.debug("\tvalue evaluated to %s", value)
            if sub.target not in script and re.search(term, script) is None:
                logger.debug("target not found")
                continue
            elif value is not None:
                sub.executed = True
                # replace once, using regex to collect differing cases
                script = re.sub(term, str(value), script)
                # replace again using the original target, since the regex will miss it
                script = script.replace(sub.target, str(value))
                logger.debug("replaced")
            elif empty_treatment in ["wipe", "local"]:  # remove nonvalued entries
                script = re.sub(term, "~mark_for_deletion~", script)
                script = script.replace(sub.target, "~mark_for_deletion~")
                logger.debug("no value, replaced with deletion mark")
        # remove any lines marked for deletion by appending those without the mark
        clean = []
        for line in script.split("\n"):
            if "~mark_for_deletion~" not in line:
                clean.append(line)
            elif empty_treatment == "local":
                tmp = line.replace("~mark_for_deletion~", "")
                clean.append(tmp)
            elif "#SUBMISSION_SUBSTITUTION#" in line:
                clean.append("#SUBMISSION_SUBSTITUTION#")
        script = "\n".join(clean)
        del clean

        return script

    def pack(self, file=None):
        if file is not None:
            self.to_yaml(filepath=file)
            return
        return self.to_dict()

    @classmethod
    def unpack(cls, data: dict = None, file: str = None, limit: bool = True):
        if file is not None:
            return cls.from_yaml(file)
        return cls.from_dict(data)


def unpack_parser(parser_data) -> Function:
    """
    Parser data can be either a string, dict or callable.
    Handle all and return the Function

    Args:
        parser_data:
            stored parser
    Returns:
        Function
    """
    if callable(parser_data):
        return Function(parser_data, force_self=True)
    if isinstance(parser_data, str):
        return Function(parser_data, force_self=True)
    if isinstance(parser_data, dict):
        return Function.unpack(parser_data, force_self=True)
    if isinstance(parser_data, Function):
        return parser_data
    return parser_data


def legacy_unpack(computer: BaseComputer, payload: dict) -> None:
    """
    Handle legacy style dictionary specs

    Args:
        computer:
            BaseComputer subclass to be updated
        payload:
            spec payload

    Returns:
        None
    """
    optional = payload.pop("optional_resources", {})
    for name, flag in optional.items():
        resource = Resource(name=name, flag=flag)
        object.__setattr__(computer, name, resource)

    required_or = payload.pop("required_or", [])
    for group in required_or:
        for name, flag in group.items():
            replaces = [n for n in group if n != name]
            resource = Resource(name=name, flag=flag, optional=False, replaces=replaces)
            object.__setattr__(computer, name, resource)

    defaults = payload.pop("optional_defaults")
    for name, default in defaults.items():
        computer.resource_dict[name].default = default

    print(
        "WARNING! Old style import detected. "
        "You should check the validity of the "
        "resulting Computer then re-dump to yaml."
    )



"""
This module handles the base functionality for serialising and unserialising
objects.

The SendableMixin class provides the necessary methods to convert objects to
YAML format and vice versa.
"""

import collections.abc
import importlib
import json
import logging
import os.path
import re
from collections import deque
from typing import Dict, List, Union

import yaml

INTERNAL_STORAGE_KEYS = {
    "CLASS_STORAGE_KEY": "~serialisedclass~",
    "TUPLE_STORAGE_KEY": "~serialisedtuple~",
    "SET_STORAGE_KEY": "~serialisedset~",
    "DEQUE_STORAGE_KEY": "~serialiseddeque~",
    "SERIALISED_STORAGE_KEY": "~serialisedobject~",
    "SLOTS_ENABLE_KEY": "~usesslots~",
}
# if modules are moved, we can exchange the name before unserialising
COMPATIBILITY = {r"remotemanager\.logging\b": "remotemanager.logging_utils"}

logger = logging.getLogger(__name__)


class SendableMixin:
    """
    Mixin class to create "sendable" object.
    Provides methods for conversion to yaml format

    Create a sendable object by subclassing SendableMixin at class creation

    >>> class MyObject(SendableMixin):
    >>>     ...

    Instances of this object will now have the required methods to be
    converted to and from dict format

    >>> new = MyObject()
    >>> payload = new.pack()  # store the object in a dict object
    >>> recreated = MyObject.unpack(payload)  # create an instance from dict
    """

    __slots__ = ["_unpack_validate"]
    serialiser = NotImplemented

    def pack(self, uuid: str = None, file: str = None) -> Union[dict, None]:
        """
        "packs" the object into a dict-format, ready for yaml-dump

        Args:
            uuid (str):
                Optional uuid string to package this data inside.
                Adds a toplevel `uuid` to the payload dict:
                >>> p = {...}

                Becomes:
                >>> p = {uuid: {...}}

            file (str):
                Directly package to file with `yaml.dump`

        Returns:
            (dict):
                object payload
        """
        # remove any attributes not to be packaged
        # these objects are either not needed, or can be re-created on the
        # recipient end
        never_package = ["_logger", "_logobj"]
        class_storage = get_class_storage(self)

        # if the object is a subclass, find the name of the parent by getting
        # the mro of the object, and extracting it from the set
        # parent_class_names = get_mro_classnames(self)

        # if the class specifies a solo _do_not_package, add that
        never_package += getattr(self.__class__, "_do_not_package", [])

        payload = {}
        if hasattr(self, "__dict__"):
            # print("standard pack")
            # caching the keys sidesteps a weird bug with serialdill and a complex
            # object where __dict__ would change size during a pack
            keys = [k for k in list(self.__dict__.keys())]
            for k in keys:
                if k in never_package:
                    # print(f"skipping k = {k}")
                    continue
                newv = self.__dict__.get(k)
                payload[self.serialise(k)] = self.serialise(newv)
        else:
            # print("slots pack")
            payload[INTERNAL_STORAGE_KEYS["SLOTS_ENABLE_KEY"]] = True
            for key in self.__slots__:
                if key in never_package:
                    continue
                payload[self.serialise(key)] = getattr(self, key, None)

        payload[INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"]] = class_storage

        if uuid:
            if "_uuid" in payload and payload["_uuid"] != uuid:
                raise ValueError("passed uuid and _uuid key in payload do not match!")
            payload = {uuid: payload}

        if file is not None:
            print(f"dumping payload to {file}")
            with open(file, "w+", encoding="utf-8") as o:
                yaml.dump(payload, o)

            return None

        return payload

    @classmethod
    def unpack(cls, data: dict = None, file: str = None, **kwargs):
        """
        Re-create an object from a packaged payload coming from ``obj.pack``

        .. note ::
            use this function to unpack from a payload _outside_ an object

            .. code:: python

                newobj = MyObject.unpack(payload)

            Where ``MyObject`` is a subclass of SendableMixin,
            and ``payload`` is a dict-type coming from ``MyObject.pack()``

        Args:
            data (dict):
                __dict__ payload from the object that was packaged
            file (str):
                filepath to unpack from, if data is not given
            limit (bool):
                set False to allow outside classes to be unserialised

        Returns:
            re-created object
        """

        if data is None:
            if file is None:
                raise ValueError(
                    "please provide a data payload or filepath to read from"
                )
            elif not os.path.isfile(file):
                raise FileNotFoundError(f"could not find file {file}")

            with open(file, "r", encoding="utf-8") as o:
                data = yaml.safe_load(o)

        # if the data passed is a {uuid: {data}} format dict, search for a ``uuid``
        # within and extract the {data}
        try:
            firstkey = list(data)[0]
            if (
                isinstance(data[firstkey], dict)
                and "_uuid" in data[firstkey]
                and data[firstkey]["_uuid"] == firstkey
            ):
                data = data[firstkey]
        except IndexError:
            # if there's no keys in data, then we definitely don't have a ``uuid``
            pass

        # create a new instance of this class
        new = cls.__new__(cls, **kwargs)

        # this is REQUIRED to be separate to the return call
        new.inject_payload(data)

        return new

    def inject_payload(self, payload: dict):
        """
        inject payload into the __dict__, effectively re-creating the object

        .. note ::
            use this function to unpack _within_ an object

            >>> class MyObject(SendableMixin):
            >>>     def __init__(self, ...):
            >>>         ...
            >>>         self.inject_payload(payload)

        Args:
            payload (dict):
                __dict__ payload from the object that was packaged
        """
        if hasattr(self, "_logger"):
            logger.info("finalising unpacking of %s", type(self))

        # temporary attribute to check for a valid class
        self._unpack_validate = True

        if INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"] in payload:
            selfkey = get_class_storage(self)
            packkey = payload[INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"]]

            if selfkey != packkey:
                raise RuntimeError(
                    f"attempting to unpack class "
                    f'{packkey["name"]} as '
                    f'{selfkey["name"]}'
                )

            delattr(self, "_unpack_validate")

        for k, v in payload.items():
            if k in list(INTERNAL_STORAGE_KEYS.values()):
                continue
            setattr(self, k, self.unserialise(v))

    def serialise(self, obj):
        """
        Recurse over any iterable objects, or call the pack() method of any
        `SendableMixin` objects, for serialisation

        Args:
            obj:
                object to be packaged

        Returns (yaml-serialisable object):
            yaml-friendly object
        """

        if issubclass(type(obj), SendableMixin):
            payload = obj.pack()
            return payload

        elif isinstance(obj, collections.abc.Mapping):
            # dict type
            try:
                payload = {self.serialise(k): self.serialise(v) for k, v in obj.items()}
                return payload
            except TypeError as e:
                raise ValueError(
                    f"Possible dictionary as key while serialising {obj}. "
                    "This can cause circular issues with serialisation\n"
                    f"Original error: {e}"
                ) from e

        elif isinstance(obj, list):
            # list-type
            return [self.serialise(v) for v in obj]

        elif isinstance(obj, tuple):
            # tuple-type
            return [INTERNAL_STORAGE_KEYS["TUPLE_STORAGE_KEY"]] + [
                self.serialise(v) for v in obj
            ]

        elif isinstance(obj, set):
            # set-type
            return [INTERNAL_STORAGE_KEYS["SET_STORAGE_KEY"]] + [
                self.serialise(v) for v in obj
            ]

        elif isinstance(obj, deque):
            # collections deque
            return [INTERNAL_STORAGE_KEYS["DEQUE_STORAGE_KEY"], obj.maxlen] + [
                self.serialise(v) for v in obj
            ]

        if basic_available(obj):
            return obj

        try:
            return [
                INTERNAL_STORAGE_KEYS["SERIALISED_STORAGE_KEY"],
                self.serialiser.dumps(obj),
            ]
        except AttributeError:
            pass

        raise RuntimeError(f"Failed to yaml dump {obj.__class__.__name__} object {obj}")

    def unserialise(self, obj):
        """
        Undo a serialised packaging, by importing the called object and calling
        its unpack() method

        Args:
            obj:
                payload to be unpacked

        Returns:
            object before packaging
        """
        try:
            unpackable = obj.get(INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"])
        except AttributeError:
            unpackable = False

        if unpackable:
            # extract the class to import
            source = obj.pop(INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"])

            # import the module
            modulename = source["mod"]

            for term, new in COMPATIBILITY.items():
                modulename = re.sub(term, new, modulename)

            # Protection is currently DISABLED
            # BaseComputer unpacks as a dict, and is therefore impossible to type check
            # regardless, if the security is disabled by simply subclassing URL, then
            # there's not much point to it in the first place, is there?
            # if limit and not modulename.startswith("remotemanager"):
            #     raise ValueError(
            #         "module to import is not within the "
            #         "remotemanager package, exiting for safety"
            #     )
            mod = importlib.import_module(modulename)

            # now get the actual class to import and unpack
            try:
                cls = getattr(mod, source["name"])
                return cls.unpack(obj)
            # this object was marked as unpackable, but the unpacking failed for
            # some reason. It's best to return None here, as returning the dict
            # can create some strange behaviours.
            except AttributeError:
                return None

        elif isinstance(obj, collections.abc.Mapping):
            # dict type
            return {self.unserialise(k): self.unserialise(v) for k, v in obj.items()}

        # coming from the yaml file, output should _only_ be list
        elif isinstance(obj, list):
            try:
                if obj[0] == INTERNAL_STORAGE_KEYS["TUPLE_STORAGE_KEY"]:
                    return tuple([self.unserialise(o) for o in obj[1:]])
                elif obj[0] == INTERNAL_STORAGE_KEYS["SET_STORAGE_KEY"]:
                    return set([self.unserialise(o) for o in obj[1:]])
                elif obj[0] == INTERNAL_STORAGE_KEYS["DEQUE_STORAGE_KEY"]:
                    maxlen = obj[1]
                    dqobj = obj[2:]
                    return deque([self.unserialise(o) for o in dqobj], maxlen=maxlen)
                else:
                    return [self.unserialise(v) for v in obj]
            except IndexError:
                return [self.unserialise(v) for v in obj]

        if basic_available(obj):
            return obj

        return obj

    def is_missing(self, objname: str) -> bool:
        """
        Determine if object with name objname is missing or uninitialised

        Args:
            objname (str): name of the object to look for

        Returns:
            (bool): object presence
        """

        if not hasattr(self, objname):
            return True

        obj = getattr(self, objname)
        if obj is None:
            return True

        return False


def get_class_storage(obj) -> Dict[str, str]:
    """
    Breaks down object into its module and classname.

    Args:
        obj:
            Python object to be broken down

    Returns (dict):
        module and classname dict
    """
    return {"mod": obj.__module__, "name": obj.__class__.__name__}


def get_mro_classnames(obj) -> List[str]:
    """
    Retrieves a list of class names from the Method Resolution Order (MRO) of an object.

    Args:
        obj: Python object whose MRO is to be retrieved.

    Returns:
        List[str]: A list containing the class names in the order they appear
        in the MRO.
    """
    return [subobj.__name__ for subobj in obj.__class__.__mro__]


def basic_available(obj) -> bool:
    """attempt a basic JSON serialisation, which will fail fast if we can't dump"""
    try:
        json.dumps(obj)
        return True
    except Exception as ex:  # pylint: disable=W0718
        if hasattr(obj, "_logger"):
            obj._logger.error(str(ex))
        return False



"""Mixin class to provide underlying verbose properties"""

from typing import Union

from remotemanager.logging_utils.verbosity import Verbosity


# pylint: disable=protected-access
def make_verbose(cls):
    """
    Adds the correct verbose properties to the class cls

    Args:
        cls:
            class to treat

    Returns:
        cls, modified
    """

    def get_verbose(self) -> Verbosity:
        """
        Return the current verbosity setting

        Returns:
            (Verbosity): current verbosity
        """
        if not isinstance(self._verbose, Verbosity):
            self._verbose = Verbosity(self._verbose)
        return self._verbose

    def set_verbose(self, verbose: Union[None, int, bool, Verbosity]) -> None:
        """
        Verbosity setter
        """
        self._verbose = Verbosity(verbose)

    cls._verbose = Verbosity()

    prop = property(fset=set_verbose, fget=get_verbose, doc="Verbose property")
    cls.verbose = prop

    return cls



import json
import os
import re
from datetime import datetime, timezone

date_format = "%Y-%m-%d %H:%M:%S %Z"
manifest_filename = "{manifest_filename}"


class Manifest:
    def __init__(
        self, instance_uuid, path_override: str = None, name_override: str = None
    ):
        self.instance_uuid = instance_uuid

        self.filename = name_override or manifest_filename

        path = path_override or __file__
        if "." in path:
            path = os.path.split(path)[0]
        self.basepath = path

        self.runner_mode = False  # set True by runners, disables printing

    def now(self) -> str:
        return datetime.strftime(datetime.now(timezone.utc), date_format)

    def to_timestamp(self, timestring: str) -> int:
        if "UTC" in timestring:
            timestring = timestring.replace("UTC", "+0000")

        try:
            dt = datetime.strptime(timestring, date_format)
        except ValueError:
            try:
                dt = datetime.strptime(timestring, "%Y-%m-%d %H:%M:%S %z")
            except ValueError:
                dt = datetime.strptime(timestring, "%Y-%m-%d %H:%M:%S")

        return int(dt.timestamp())

    @property
    def path(self):
        return os.path.join(self.basepath, self.filename)

    def get_data(self, path: str = None) -> list:
        if path is None:
            path = self.path
        if not os.path.exists(path):
            return []
        with open(path) as o:
            data = o.readlines()
        return data

    def write(self, string: str):
        string = f"{self.now()} [{self.instance_uuid}] {string.strip()}\n"
        with open(self.path, "a+") as o:
            o.write(string)

    def get(self, uuid: str) -> list:
        """Gets all lines that correspond to uuid"""
        # we need to "recreate" the log from the parsed output,
        # since parse_log handles multiline entries
        log = []
        for entry in self.parse_log()["log"].get(uuid, []):
            # the timestamp here is always from a UTC string
            dt = datetime.fromtimestamp(entry[0]).astimezone(timezone.utc)
            timestring = dt.strftime("%Y-%m-%d %H:%M:%S UTC")
            log.append(f"{timestring} [uuid_a] {entry[1]}")

        return log

    def parse_log(
        self, path: str = None, string: str = None, cutoff: int = None
    ) -> dict:
        """
        Convert the log into a dict of {uuid: [time, log]}

        Args:
            path: Override filepath
            string: Override file read with string input
            cutoff: Only consider events after this timestamp
        """
        if string is not None:
            data = string.split("\n")
        else:
            data = self.get_data(path)

        def add_to_output(output: dict, uuid: str, timestamp: int, cache: list) -> None:
            if len(cache) == 0:
                return
            content = "\n".join(cache)

            add_to = "log"
            # first split off stdout/stderr section using maxsplit=1
            # then take the last section with [-1]
            # this leaves an additional single space that should be dropped with [1:]
            # strip is dangerous, since it drops _all_ whitespace
            if content.startswith("[stdout]"):
                add_to = "stdout"
                content = content.rsplit("[stdout]", maxsplit=1)[-1][1:]
            elif content.startswith("[stderr]"):
                add_to = "stderr"
                content = content.rsplit("[stderr]", maxsplit=1)[-1][1:]

            try:
                output[add_to][uuid].append((timestamp, content.rstrip()))
            except KeyError:
                output[add_to][uuid] = [(timestamp, content.rstrip())]

        date_regex = re.compile(r"\d{4}-\d{2}-\d{2}")

        cache = []
        output = {"log": {}, "stdout": {}, "stderr": {}}

        uuid = None
        ts = 0
        for line in data:
            if re.match(date_regex, line):
                add_to_output(output, uuid, ts, cache)
                cache = []

                ts = self.to_timestamp(line[: line.index("[")].strip())

                if cutoff is not None and cutoff > ts:
                    continue

                uuid_index = line.index("]")

                uuid = line[25:uuid_index]
                # fmt: off
                content = line[uuid_index + 2:]
                # fmt: on

                cache.append(content.strip())

            else:  # continuation of a previous line, store it for later
                cache.append(line.strip())

        if len(cache) != 0:
            add_to_output(output, uuid, ts, cache)

        return output

    def last_time(self, state: str) -> dict:
        """Takes parsed log, returning the last time state appeared for each uuid"""
        data = self.parse_log()["log"]
        output = {}
        for uuid, log in data.items():
            if uuid not in output:
                output[uuid] = None
            for line in log:
                if line[1].strip().lower() == state.lower():
                    output[uuid] = line[0]

        if not self.runner_mode:
            print(json.dumps(output))
        return output


# DATASET_CONTENT #



import logging
import warnings
from typing import Any, Union

from remotemanager.connection.computers.dynamicvalue import DynamicMixin

logger = logging.getLogger(__name__)


class Substitution(DynamicMixin):
    """
    Stores a jobscript template substitution

    Args:
        target:
            String to _replace_ in the template.
            Ideally should begin with a commenting character (#)
        name:
            String to replace _with_. At script generation, `target`
            will be replaced with `name`
        mode:
            Used by JUBETemplate

    .. note::
        For other args see the DynamicMixin class
    """

    __slots__ = ["target", "mode", "executed", "dependencies"]

    def __init__(self, target: str, name: str, mode: Union[str, None] = None, **kwargs):
        super().__init__(assignment=name, **kwargs)

        self.target = target
        self.mode = mode

        self.executed = False
        self.dependencies = []  # must calculate the value of these first

    def __hash__(self) -> int:
        return hash(self.target + self.name)

    def __str__(self):
        return str(self.value)

    def __repr__(self) -> str:
        return f"Substitution({self.target}, {self.name}) -> {self.value}"

    def __bool__(self):
        """
        Makes objects "falsy" if no value has been set, "truthy" otherwise
        """
        return self.value is not None

    @classmethod
    def from_string(
        cls, string: str, warn_invalid: bool = True, **kwargs
    ) -> "Substitution":
        """
        Create a substitution object from template string

        Args:
            string (str): Input string to generate from
            warn_invalid (bool): Invalid args name, target will be deleted. Warn if True
            kwargs: Any keyword args to override the string with
        """

        content = string.strip("#")  # actual internal "content"
        sym = content.split(":")[0]  # name, drop args for function extract
        string_args = cls.get_target_kwargs(string)

        logger.debug("\ttarget=%s, symbol=%s, args=%s", string, sym, string_args)

        invalid_args = ["name", "target"]
        for arg in invalid_args:
            if arg in kwargs:
                if warn_invalid:
                    warnings.warn(f"Invalid kwarg {arg}={kwargs[arg]} deleted")
                del kwargs[arg]

        string_args.update(kwargs)

        name = sym.lower()

        logger.debug("\tprocessed symbol: %s", string)
        return cls(name=name, target=string, **string_args)

    @property
    def target_kwargs(self):
        """Attempts to extract the kwargs from the target string"""
        return self.get_target_kwargs(self.target)

    @staticmethod
    def get_target_kwargs(string: str) -> dict:
        """Attempts to generate kwargs from input string"""

        if ":" not in string:
            return {}

        _, argline = string.strip("#").split(":", maxsplit=1)

        args = {}
        key_cache = []
        val_cache = []
        key_mode = True
        escape = False
        inset_cache = []

        def append(c: str):
            if key_mode:
                key_cache.append(c)
            else:
                val_cache.append(c)

        for char in argline:
            # handle escaped characters
            if char == "\\" and not escape:
                escape = True
                append(char)
                continue
            # we're in escape mode, append the char, set the flag false and continue
            if escape:
                escape = False
                append(char)
                continue

            # swap from arg to val, once
            if char == "=" and key_mode:
                key_mode = False
                continue

            # quotation
            if char == '"':
                if len(inset_cache) == 0:
                    inset_cache.append(char)
                elif inset_cache[-1] == '"':
                    inset_cache.pop(-1)
            if char == "'":
                if len(inset_cache) == 0:
                    inset_cache.append(char)
                elif inset_cache[-1] == "'":
                    inset_cache.pop(-1)

            # evaluation
            if char == "{":
                inset_cache.append("{")
            if char == "}":
                if inset_cache[-1] == "{":
                    inset_cache.pop(-1)

            # end of this arg, reset
            if char == ":" and len(inset_cache) == 0:
                if key_mode:
                    raise ValueError(
                        f"Spurious ':' character in arg {''.join(key_cache)}"
                    )

                args["".join(key_cache)] = "".join(val_cache)
                key_mode = True
                key_cache = []
                val_cache = []
                continue

            # otherwise, store this char and continue
            append(char)

        if len(key_cache) != 0 and len(val_cache) != 0:
            args["".join(key_cache)] = "".join(val_cache)

        return args

    @property
    def value(self) -> Any:
        """
        Returns:
            value if present, else default
        """
        val = super().value

        if len(self.dependencies) == 0:
            return val
        # if we have dependents, we need to evaluate them
        evaluate = {}
        # int values cause this replace to fail
        # non strings probably shouldn't end up here,
        # but it's best not to error weirdly
        val = str(val)
        for dep in self.dependencies:
            evaluate[dep.arg] = dep.value
            # replace the target $value with the evaluated arg
            # print(f"replacing {dep.arg} -> {dep.value}")
            val = val.replace(f"${dep.arg}", str(dep.value))
        # now everything is replaced, we can freely eval
        # print(f"evaluating {val}")
        val = eval(str(val.strip()))
        # print(f"{self.arg}={val}\n")
        # cache the result
        self.value = val
        self.dependencies = {}

        return val

    @value.setter
    def value(self, value):
        super().set_value(value)

    @property
    def arg(self) -> str:
        """
        Returns:
            Argument which is exposed to the underlying URL for setting
        """
        return self.name.strip("$")

    @property
    def entrypoint(self) -> str:
        """Returns the name/name for this sub"""
        return self.name

    def pack(self, collect_value: bool = True) -> dict:
        """Store this Substitution in dict form"""
        data = super().pack(collect_value=collect_value)

        if self.mode is not None:
            data["mode"] = self.mode
        if self.target is not None:
            data["target"] = self.target

        return data



"""
Handles file transfer via the `rsync` protocol
"""

import logging
import os.path
import pathlib
import re

from remotemanager.transport.transport import Transport
from remotemanager.utils.version import Version

logger = logging.getLogger(__name__)

VERSION_REGEX = r"version.(\d\.\d\.\d)"
RSYNC_MIN_VERSION = "3.0.0"
TRANSPORT_CHANGE_DOCS_URL = (
    "https://l_sim.gitlab.io/remotemanager/tutorials/E7_Changing_Transport.html"
)


class rsync(Transport):
    """
    Class for `rsync` protocol

    Args:
        checksum (bool):
            Adds checksum arg, which if ``True`` will add ``--checksum`` flag to
            parameters
        progress (bool):
            request streaming of rsync --progress stdout
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_version()

        # flags can be exposed, to utilise their flexibility
        flags = kwargs.pop("flags", "auvh")
        self.flags = flags

        if kwargs.get("checksum", True):
            self.flags += "--checksum"

        if "progress" in kwargs and kwargs["progress"]:
            logger.debug("rsync progress requested")
            self.flags += "--progress"
            self._request_stream = True

        logger.info("created new rsync transport")

    def check_version(
        self, collect_only: bool = False, min_version: str = None
    ) -> Version:
        """
        Queries the installed rsync version and checks that it is recent enough

        Does nothing if the version could not be detected, so should be considered a
        "soft" check, rather than a true safety feature.

        Args:
            collect_only (bool):
                Does not perform a check if True. Defaults to False
            min_version (str):
                override the minimum version (used for testing)

        Returns:
            Version
        """
        if min_version is not None:
            print(
                f"Warning: rsync check_version is being run "
                f"with a modified min version: {min_version}"
            )
        else:
            min_version = RSYNC_MIN_VERSION

        content = self.url.cmd("rsync --version", local=True).stdout
        search = re.search(VERSION_REGEX, content)

        if search is None:
            return Version("0.0.0")

        version = Version(search.group(1))

        if not collect_only and version < Version(min_version):
            raise RuntimeError(
                f"rsync version ({version}) is less than the required {min_version}\n"
                "Please update your install, or swap to a different Transport\n"
                "More info can be found here: "
                f"{TRANSPORT_CHANGE_DOCS_URL}"  # noqa: E501
            )

        return version

    def cmd(self, primary, secondary):
        if self.url.passfile and self.url.keyfile:
            raise RuntimeError(
                "rsync appears to have an issue when "
                "specifying sshpass AND ssh-key. Either set up "
                "your ssh config and remove the keyfile or use "
                "transport.scp"
            )

        password = ""
        if self.url.passfile is not None:
            password = f'--rsh="{self.url.passfile} ssh" '

        insert = ""
        if self.url.ssh_insert != "":
            insert = f'-e "ssh {self.url.ssh_insert}" '

        cmd = "rsync {flags} {ssh_insert}{password}{inner_dir}{primary} {secondary}"
        inner_dir = ""
        if len(pathlib.Path(secondary).parts) > 1:
            # the target is a nested dir. If the whole tree doesn't exist,
            # rsync will throw an error
            if ":" in secondary:
                # target is a remote folder, use the --rsync-path hack
                inner_dir = (
                    f'--rsync-path="mkdir -p '
                    f'{Transport.get_remote_dir(secondary)} && rsync" '
                )
            elif not os.path.exists(secondary):
                os.makedirs(secondary)

        base = cmd.format(
            flags=self.flags,
            ssh_insert=insert,
            password=password,
            primary=primary,
            secondary=secondary,
            inner_dir=inner_dir,
        )
        logger.debug(f'returning formatted cmd: "{base}"')
        return base



"""
This module stores the placeholder arguments who's job it is to convert
arguments from the Dataset level `mpi`,  `omp`, `nodes`, etc. to what the
scheduler is expecting within a jobscript.

.. note::
    Placeholders without a `value` are "falsy". So checking their value in an
    if statement will return `True` if they have a value, `False` otherwise.
"""

import logging
from typing import Union

from remotemanager.connection.computers.dynamicvalue import DynamicMixin

logger = logging.getLogger(__name__)


class Resource(DynamicMixin):
    """
    Stub class to sit in place of an option within a computer.

    Args:
        name (str):
            name under which this arg is stored
        flag (str):
            Flag to append value to e.g. `--nodes`, `--walltime`
        separator (str):
            Override the separator between flag and value (defaults to "=")
        tag (str):
            Override the tag preceding the flag (defaults to "--")

    .. note::
        For other args see the DynamicMixin class
    """

    __slots__ = [
        "flag",
        "pragma",
        "tag",
        "separator",
    ]

    def __init__(
        self,
        name: str,
        flag: Union[str, None] = None,
        tag: Union[str, None] = None,
        separator: Union[str, None] = None,
        **kwargs,
    ):
        super().__init__(assignment=name, **kwargs)
        self.flag = flag

        self.pragma = None
        self.tag = tag
        self.separator = separator

    def __hash__(self):
        return hash(self.flag)

    def __repr__(self):
        return str(self.value)

    def __bool__(self):
        """
        Makes objects "falsy" if no value has been set, "truthy" otherwise
        """
        return self.value is not None and self.flag is not None

    @property
    def resource_line(self) -> str:
        """
        Shortcut to output a suitable resource request line

        Returns:
            str: resource request line
        """
        pragma = f"{self.pragma} " if self.pragma is not None else ""
        tag = self.tag if self.tag is not None else "--"
        separator = self.separator if self.separator is not None else "="

        return f"{pragma}{tag}{self.flag}{separator}{self.value}"

    def pack(self, collect_value: bool = True) -> dict:
        """Store this Resource in dict form"""
        data = super().pack(collect_value=collect_value)

        for k in self.__slots__:
            v = getattr(self, k, None)

            if v is not None:
                data[k] = v

        return data


class runargs(dict):
    """
    Class to contain the dataset run_args in a way that won't break any loops
    over the resources

    Args:
        args (dict):
            Dataset run_args
    """

    _accesserror = (
        "\nParser is attempting to access the flag of the run_args, you "
        "should add an `if {option}: ...` catch to your parser."
        "\nRemember that placeholders without an argument are 'falsy', "
        "see the docs for more info. https://l_sim.gitlab.io/remotemanager"
        "/remotemanager.connection.computers.options.html"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __bool__(self):
        return False

    @property
    def value(self):
        """
        Prevents an AttributeError when a parser attempts to access the value.

        Returns:
            (dict): internal dict
        """
        return self.__dict__

    @property
    def flag(self):
        """
        Parsers should not access the flag method of the run_args, doing so likely
        means that a loop has iterated over this object and is attempting to insert
        it into a jobscript.

        Converts an AttributeError to one more tailored to the situation.

        Returns:
            RuntimeError
        """
        raise RuntimeError(runargs._accesserror)


class Resources:
    """
    Container class to store Resource objects for use by a parser
    """

    __slots__ = ["_names", "_resources", "_run_args", "pragma"]

    def __init__(self, resources, pragma, tag, separator, run_args):
        self._names = []
        self._resources = resources
        self._run_args = run_args

        self.pragma = pragma

        for resource in self._resources:
            self._names.append(resource.name)
            # add pragma to Resource for resource_line property
            resource.pragma = pragma
            if resource.tag is None:
                resource.tag = tag
            if resource.separator is None:
                resource.separator = separator

    def __iter__(self):
        return iter(self._resources)

    def __getitem__(self, item: str) -> Union[Resource, dict]:
        """
        Need to enable Resources["mpi"], for example

        Args:
            item:
                name of resource to get

        Returns:
            Resource
        """
        if item == "run_args":
            return self.run_args
        try:
            return self._resources[self._names.index(item)]
        except ValueError:
            raise ValueError(f"{self} has no resource {item}")

    def get(self, name: str, default: any = "_unspecified"):
        """Allows resource.get(name)"""
        if default == "_unspecified":
            return getattr(self, name)
        return getattr(self, name, default)

    def items(self):
        """dict.items() like proxy"""
        for resource in self:
            yield resource.name, resource

    @property
    def run_args(self) -> dict:
        """Returns the stored run_args"""
        return self._run_args
